#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
SYNOPSIS
    preprocessing_pm10.py -i <folder_path> -o <folder_path> -y <year> [-v,--verbose] [-h,--help] [--version]

DESCRIPTION
    This snippet is for preprocessing of the pm10 data from https://ads.atmosphere.copernicus.eu/cdsapp#!/dataset/cams-global-reanalysis-eac4-monthly?tab=form.
    The average pm10 concentration out of all input data in a folder. Then the raster is
    resampled plus warped to the needed AOI and resolution.


PREREQUISITES
    Python => 3.7.0

EXAMPLES
    python -m inca.airfiltration.preprocessing_pm10 -i <folder_path> -o <folder_path> -y <year>
    OR
    python3 preprocessing_pm10.py -i <folder_path> -o <folder_path> -y <year>

EXIT STATUS
    Ctrl-C  =  cancel the script at any time

AUTHOR
    Dr. Bert De Roo <bert.deroo@vito.be>
"""
################## Importing modules ##################
# standard
import logging
import optparse
import os
import time
import subprocess
from calendar import monthrange
from datetime import datetime, timedelta, date

# extra official modules
import rasterio
import numpy as np

# own modules
import inca
import inca.common.geoprocessing as geo

############## END importing modules ##################

logger = logging.getLogger(__name__)


############## MAIN ###################################
def main():
    start_time = time.time()
    # parse the arguments
    options = parse_args()

    # ini the logging
    inca.set_up_console_logging(options.verbose)

    # change umask driver
    try:
        old_umask = os.umask(0o000)
    except Exception:
        logger.exception('Could not change UMASK setting')
        pass

    logger.info('START OF MODULE: INCA air filtration preprocessing')

    # init the metadata and raster extent of project
    metadata = geo.Metadata("Dr. Bert De Roo for INCA project",
                            "air filtration pre-processing"
                            )
    os.makedirs(os.path.normpath(options.path_out), exist_ok=True)

    # run the total pm10 generation
    logger.info('generating the total pm10 in original resolution')
    try:
        processing_total_pm10(options, metadata)
    except Exception:
        logger.exception('Error during processing of the datasets to generate the total pm10')
        raise



    logger.info('All processing successfully done. Have a nice day.')
    # set umask to old settings
    try:
        os.umask(old_umask)
    except Exception:
        logger.exception('!!! Could not change UMASK setting back!!!')
        pass

    logger.debug('---------------------------')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: INCA air filtration PM10 preprocessing')

def parse_args():
    parser = optparse.OptionParser(formatter=optparse.TitledHelpFormatter(), usage=globals()['__doc__'],
                                   version="%prog v1.0")
    parser.add_option('-v', '--verbose', action='store_true', default=False, help='verbose output')
    parser.add_option('-i', '--path_in', help='Path to folder with PM10 input data. Needed.', dest='path_in')
    parser.add_option('-o', '--path_out', help='Path to output folder. Needed.', dest='path_out')
    parser.add_option('-y', '--year', help='Processing year. Needed.', dest='pYear')

    # parse the given system arguments
    (options, args) = parser.parse_args()
    # do checks on the parsed options
    if (not options.path_in) or (not os.path.isdir(os.path.normpath(options.path_in))):
        parser.error("the -i argument for PM10 input data is missing or the given path doesn't exist!!!")
    if (not options.pYear) or (not options.pYear.isdigit()):
        parser.error("the -y argument must be a year.")

    if len(args) != 0:
        parser.error('too many arguments')

    return options

def output(outname, data, src_profile, options, metadata):
    '''
    writes out the data to outname afterwards cuts it to aoi
    :param outname: name of the output file
    :param data: data to write
    :param src_profile: rasterio profile to be used to write the data
    :param options: options
    :param metadata: metadata object from geoprocessing that needs be infused
    :return:
    '''

    #setup dst profile
    src_profile.update(count=1, dtype=rasterio.int16, driver= 'gtiff')

    #output filename orginal epsg
    original_proj_path  = os.path.dirname(os.path.join(os.path.normpath(options.path_out).replace('input_model','input_intermediate'), outname))
    if not os.path.exists(original_proj_path):
        os.makedirs(original_proj_path)

    path_out = os.path.join(os.path.normpath(options.path_out), outname + '_pm10_average.tif'.format(options.pYear))
    path_inter = path_out.replace('input_model','input_intermediate')

    tags = metadata.prepare_raster_tags(
        'Average pm10 derived from CGLS datasets for the year {}. value_range = [0, ...]. scaling_factor: {}'.format(
            options.pYear, options.scale_factor), 'pm10')

    try:
        with rasterio.open(path_inter, 'w', **src_profile, compress='DEFLATE') as dst:
            dst.scales = [options.scale_factor]
            dst.offsets = [options.offset]
            dst.update_tags(**tags)
            dst.update_tags(1, scale_factor=options.scale_factor, offset = options.offset)
            dst.write(data.astype(src_profile['dtype']), 1)
    except:
        if os.path.exists(path_out):
            os.remove(path_out)
        raise IOError('generation of average PM10 for year {} was not successful'.format(options.pYear))

    #bring_2_AOI(options, metadata, path_inter, path_out)

def bring_2_AOI(options, metadata, path_inter ,path):
    '''
    function of warp original dataset to correct EPSG, resolution and AOI
    Note: I'm not using the function from the geoprocessing.py due to needed adaptations

    :param options: options
    :param metadata: metadata object from geoprocessing that needs be infused
    :param path_inter: file that needs to be warped to correct EPSG, resolution and AOI
    :param path: warped file
    '''

    #final path_out
    path_out = path.replace('.tif','_INCA_{}m_EPSG{}.tif'.format(
        100,
        '3035'))
    if  not os.path.exists(os.path.dirname(path_out)):
        os.makedirs(os.path.dirname(path_out))

    cmd = 'gdalwarp --config GDAL_CACHEMAX 256 -t_srs {} -te {} {} {} {} -tr {} {} -r {} -et 0 -nomd -wo SAMPLE_STEPS=50 -wo SOURCE_EXTRA=5 -wo SAMPLE-GRID=YES -co COMPRESS=DEFLATE -co TILED=YES -co INTERLEAVE=BAND -co BIGTIFF=YES -multi -overwrite {} {}'.format(
        'EPSG:3035',
        2549000,
        1386000,
        6558000,
        5447000,
        100,
        100,
        'nearest',
        path_inter,
        path_out)

    try:
        subprocess.check_call(cmd, shell=True)
    except Exception:
        raise OSError('Could not warp the raster file to the AOI')

    #infuse extra metadata
    metadata.read_raster_tags([path_inter])
    tags = metadata.prepare_raster_tags(
        'Annual average of pm10 derived from . scaling_factor: {}. Warped to EPSG, resolution and extent of INCA AOI'.format(
             options.scale_factor), 'fraction')

    with rasterio.open(path_out, 'r+') as dst:
        dst.update_tags(**tags)
        dst.update_tags(1, scale_factor= options.scale_factor, off_set= options.offset)

def processing_total_pm10(options, metadata, scale_factor = 0.01, valid_range = [-32766, 32767]):
    '''Calculate the average pm10 lai out of given files in folder for every year, quarter and month.

    NOTE: there is done no check - all tif files in folder are used to calculate total standard for CGLS 1km product.

    :param options: options
    :param metadata: metadata object from geoprocessing that needs be infused
    :param scale_factor: scale factor
    :param valid_range: valid range
    '''
    year = options.pYear
    #we do not care about memory so we run all in one chunk
    #get list of all files in input folder
    lFiles = []
    for entry in os.scandir(os.path.normpath(options.path_in)):
        if entry.is_file() and entry.name.endswith(".nc"):
            lFiles.append(entry.path)

    #Need to be sorted base on timestamp
    lFiles.sort()

    #setup tags
    metadata.read_raster_tags(lFiles)
    #get main profile data from first file
    with rasterio.open('netcdf:'+lFiles[0]+':pm10') as src:
        src_profile = src.profile
        src_profile.update({'crs':rasterio.crs.CRS.from_epsg('4326')})
        try:
            s = float(src.tags(1)['scale_factor'])
            if (s >= 0) and (s <= 1.):
                options.scale_factor = s
        except:
            logger.info('use standard scale factor since no matching metadata.')
        try:
            o = float(src.tags(1)['add_offset'])
            if (o >= 0) and (o <= 1.):
                options.offset = o
        except:
            logger.info('use standard offset factor since no matching metadata.')


        try:
            v = [int(x) for x in (src.tags(1)['valid_range']).replace(']',' ').split(' ') if x.isnumeric()]
            if (len(v) == 2) and (v[0] < v[1]):
                valid_range = v
        except:
            logger.info('use standard valid data range since no matching metadata.')

    #ini the arrays for sum and count

    aSum = np.zeros((src_profile['height'], src_profile['width']))
    aDays = np.zeros_like(aSum)

    aSum_M = aSum.copy()
    aSum_Q = aSum.copy()
    aSum_Y = aSum.copy()
    aDays_M = aDays.copy()
    aDays_Q = aDays.copy()
    aDays_Y = aDays.copy()

    maxDays_M = 0
    maxDays_Q = 0
    maxDays_Y = 0

    #no we loop over all files, files should be sorted by date and be complete (every month at least 1 data point)

        #get data
    #first getting bands?
    bands = []
    with rasterio.open('netcdf:'+ lFiles[0]+ ':pm10') as src:
        nbr_bands = src.profile['count']
        for band in range(1, 1 + nbr_bands):
            tags = src.tags(band)
            refdate  =  date(1900, 1,1)
            timeref = tags['NETCDF_DIM_time']
            time_coverage_start = refdate + timedelta(hours=int(timeref))
            if time_coverage_start.year != int(options.pYear):
                continue
            bands.extend([band])

        for band in bands:
            tags = src.tags(band)
            refdate  =  date(1900, 1,1)
            timeref = tags['NETCDF_DIM_time']
            time_coverage_start = refdate + timedelta(hours=int(timeref))
            startday,days = monthrange(int(year), time_coverage_start.month)
            aData = src.read(band)
            aData = np.ma.masked_where(aData == src_profile['nodata'],aData).astype(np.float64) \
                    *options.scale_factor + options.offset

            if band == bands[0]:
                time_coverage_start_previous = time_coverage_start
                time_coverage_start_previous_Q = time_coverage_start

            #create monthly output
            if time_coverage_start_previous.month != time_coverage_start.month:
                time_coverage_start_previous = time_coverage_start
                # invalid data is not taken into account.
                aAvg_M = (aSum_M/aDays_M-options.offset)/options.scale_factor
                aAvg_M[aDays_M==0]=src_profile['nodata']
                output(f'{year}/monthly/{str(time_coverage_start.month-1).zfill(2)}', aAvg_M, src_profile, options,metadata )
                #put in avg_q
                aSum_Q += aSum_M
                aDays_Q += aDays_M
                maxDays_Q += maxDays_M

                #reset_ months
                maxDays_M = 0
                aDays_M = aDays.copy()
                aSum_M = aSum.copy()

                #create Quarterly output
                if time_coverage_start_previous_Q.month +3 <=  time_coverage_start.month:
                    time_coverage_start_previous_Q =  time_coverage_start
                    Q = (time_coverage_start_previous_Q.month -1) //3
                    aAvg_Q = (aSum_Q/aDays_Q-options.offset)/options.scale_factor
                    aAvg_Q[aDays_Q==0]=src_profile['nodata']
                    output(f'{year}/seasonal/{str(Q).zfill(2)}', aAvg_Q, src_profile, options, metadata)
                    #put in avg_y
                    aSum_Y += aSum_Q
                    aDays_Y += aDays_Q
                    maxDays_Y += maxDays_Q
                    #reset_quarter
                    maxDays_Q = 0
                    aDays_Q = aDays.copy()
                    aSum_Q = aSum.copy()
            print(f'{time_coverage_start}')


            #get valid data
            mValid = (aData >= valid_range[0]) & (aData <= valid_range[1])
            #fill prepared arrays months
            aSum_M[mValid] += aData[mValid]*days
            aDays_M[mValid] += days
            #maxDays_M += (time_coverage_end-time_coverage_start).days


            if band == bands[-1]:
                aAvg_M = (aSum_M/aDays_M-options.offset)/options.scale_factor
                aAvg_M[aDays_M==0]=src_profile['nodata']
                output(f'{year}/monthly/{str(time_coverage_start.month).zfill(2)}', aAvg_M,  src_profile, options, metadata )
                #put in avg_q
                aSum_Q += aSum_M
                aDays_Q += aDays_M
                maxDays_Q += maxDays_M

                #create Quarterly output
                Q = 4
                aAvg_Q = (aSum_Q/aDays_Q-options.offset)/options.scale_factor
                aAvg_Q[aDays_Q==0]=src_profile['nodata']
                output(f'{year}/seasonal/{str(Q).zfill(2)}', aAvg_Q, src_profile, options, metadata)
                #put in avg_y
                aSum_Y += aSum_Q
                aDays_Y += aDays_Q
                maxDays_Y += maxDays_Q

                #create yearly output
                aAvg_Y = (aSum_Y/aDays_Y-options.offset)/options.scale_factor
                aAvg_Y[aDays_Y==0]=src_profile['nodata']
                output(f'{year}/yearly/{str(1).zfill(2)}', aAvg_Y, src_profile, options, metadata)



if __name__ == '__main__':
    # hack to get the proper logger when running the script directly
    logger = logging.getLogger('inca.airfiltration.preprocessing_pm10')
    main()
