#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
SYNOPSIS
    preprocessing_clcacc.py -i <folder_path> -o <folder_path> -y <list years> -c <coastal_mask_file> [-v,--verbose]
    [-h,--help] [--version]

DESCRIPTION
    This snippet is for preprocessing of the CLC accounting layers to prepare for ecosystem extent accounts.
    More particular it uses a coastal mask to split some CLC classes and imprint the burned area with prior class.
    The coastal-mask is provided by the European Environment Agency, as also the CLC accounting layers.

PREREQUISITES
    Python => 3.7.0

EXAMPLES
    python -m inca.common.preprocessing.preprocessing_clcacc -i <folder_path> -o <folder_path> -c <coastal_mask> -y <list of years>
    OR
    python3 preprocessing_clcacc.py -i <folder_path> -o <folder_path> -c <coastal mask> -y <list of years>

EXIT STATUS
    Ctrl-C  =  cancel the script at any time

AUTHOR
    Bruno Smets <bruno.smets@vito.be>

VERSION
    1.0 (2023-07-10)
"""

'''
IMPORTANT NOTICE
If you do not have the CoastalMask, then ...:
- TODO
'''
################## Importing modules ##################
# standard
import logging
import optparse
import os, sys
import time
from datetime import datetime, timedelta
#extra official modules
import rasterio
import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix
from tqdm import tqdm
# inca specific modules
import inca
import inca.common.geoprocessing as geo
import inca.common.preprocessing as preprocessing

############## END importing modules ##################

years = [2000, 2006, 2012, 2018]   # list of available years to trace back prior class for burn scars
block_shape = (1024, 1024)         # memory constraint used in block processing: 1M pix * uint16 = 2MB

logger = logging.getLogger(preprocessing.__name__)

############## MAIN ###################################
def main():
    start_time = time.time()
    # parse the arguments
    options = parse_args()

    # ini the logging
    inca.set_up_console_logging(options.verbose)

    # change umask driver
    try:
        old_umask = os.umask(0o000)
    except Exception:
        logger.exception('Could not change UMASK setting')
        pass

    logger.info('START OF MODULE: INCA CLCACC extent preprocessing')

    # init the metadata and raster extent of project
    accord = geo.GeoProcessing("INCA-tool",
                               "Extent (Corine Land Cover) pre-processing",
                               os.path.normpath(options.path_out),
                               ProjectExtentTiff=os.path.normpath(os.path.join(options.path_in,'CLC2000ACC_V2018_20.tif')),
                               GDAL_verbose=options.verbose)
    # make sure output folder exist
    os.makedirs(os.path.normpath(options.path_out), exist_ok=True)

    # run the total dmp generation
    logger.info('generating the adjusted Corine accounting layers (CLCACC) for year {}'.format(options.pYear))
    try:
        CLCfile = 'CLC'+str(options.pYear)+'ACC_V2018_20.tif'
        statistics = pre_process_CLCACC(options, accord, CLCfile, accord.metadata)
        statistics.to_csv(os.path.join(options.path_out, os.path.splitext(CLCfile)[0]+'_split.csv'))
        print(statistics.sum())
    except Exception:
        logger.exception('Error during processing of the datasets to preprocess CLCACC')
        raise

def pre_process_CLCACC(options, accord, CLCfile, metadata):
    # Corine standard classes
    stat_keys=[111,112,121,122,123,124,131,132,133,141,142,211,212,213,221,222,223,231,241,242, \
               243,244,311,312,313,321,322,323,324,331,332,333,334,335,411,412,421,422,423, \
               511,512,521,522,523]
    # Corine classes to split with coastal mask -> add 1000 for coastal classes
    coastal_to_split = [331,332,333]
    stat_keys.extend([x + 1000 for x in coastal_to_split])

    aStats = np.zeros((len(stat_keys),len(stat_keys)))
    with rasterio.open(os.path.join(options.path_in, CLCfile),'r') as src:
        #profile = accord.ref_profile.copy()
        profile = src.profile.copy()
        profile['tiled'] = True

    # setup tags
    metadata.read_raster_tags([os.path.join(options.path_in,CLCfile)])
    tags = metadata.prepare_raster_tags(
            f'Corine Accounting Layers prepare for Ecosystem Extent year {options.pYear}. '
            f'Using coastal mask {options.path_coastal_mask}. scaling_factor=1.0. offset_factor=0')

    with rasterio.open(os.path.join(options.path_in,CLCfile),'r') as src, \
            rasterio.open(options.path_coastal_mask,'r') as src_mask, \
            rasterio.open(os.path.join(options.path_out,os.path.splitext(CLCfile)[0]+'_split.tif'),'w',**profile) as ds_out:

        t = tqdm(total=geo.number_blocks(profile, block_shape))
        for i, dst_window in geo.block_window_generator(block_shape,profile['height'],profile['width']):
            dst_win_bounds = src.window_bounds(dst_window)
            src_window = src.window(dst_win_bounds[0],dst_win_bounds[1],dst_win_bounds[2],dst_win_bounds[3])

            # split Coastal classes according coastal mask
            aData = src.read(1, window=src_window, masked=True)
            aOrigin = aData.copy()
            aCoastalMask = src_mask.read(1, window=src_window, masked=True)
            aData = split_class(aData, aCoastalMask, classesToSplit=coastal_to_split)

            # imprint burns with class from previous year(s)
            if len((aData == 334).nonzero()[0]) > 0:
                aData = imprint_burn(options, aData, CLCfile, src_window)

            ds_out.write(aData.filled(src.nodatavals[0]), window=dst_window, indexes=1)
            aStats = get_statistics_fast(aOrigin, aData, aStats, stat_keys)
            t.update()

        ds_out.update_tags(**tags)
        ds_out.scales = [1.0]
        ds_out.offsets = [0]

        t.close()
        dStats = pd.DataFrame(aStats, columns=stat_keys, dtype = np.uint64)
        dStats.index = stat_keys

    return dStats

def imprint_burn(options, CLC, CLCfile, src_window):
    idx = years.index(int(options.pYear))
    if idx == 0:
        # assign to unassigned
        pass
    else:
        # read previous year(s)
        while idx > 0:
            prevYear = years[idx-1]
            with rasterio.open(os.path.join(options.path_in, CLCfile.replace(str(options.pYear)+'ACC',str(prevYear)+'ACC')), 'r') as src:
                aData = src.read(1, window=src_window, masked=True)

            # get current burn scars and imprint previous value
            aMask = (CLC == 334)
            if len(aMask.nonzero()[0]) > 0:
                #np.count_nonzero(aMask) > 0:
                CLC[aMask] = aData[aMask]

            if np.count_nonzero(CLC == 334):
                idx -= 1
                continue

            break

    return CLC
def split_class(CLC, inMask, classesToSplit = [331,332,333,411,412]):

    for classSplit in classesToSplit:
        mask = ((inMask == 1) & (CLC==classSplit))
        if np.ma.count(mask) > 0:
            CLC[mask] = 1000+classSplit
    return CLC

def get_statistics_fast(matrixIn, matrixOut, aStats, labels):
    if np.ma.is_masked(matrixIn):
        return aStats
    return np.add(aStats, confusion_matrix(matrixIn.flatten(),matrixOut.flatten(),labels=labels))

def parse_args():
    parser = optparse.OptionParser(formatter=optparse.TitledHelpFormatter(), usage=globals()['__doc__'],
                                   version="%prog v1.1")
    parser.add_option('-v', '--verbose', action='store_true', default=False, help='verbose output')
    parser.add_option('-c', '--path_coastal_mask', help='Path to raster coastal mask file. Needed.',
                      dest='path_coastal_mask')
    parser.add_option('-i', '--path_in', help='Path to CLCACC input files. Needed.', dest='path_in')
    parser.add_option('-o', '--path_out', help='Path to output folder. Needed.', dest='path_out')
    parser.add_option('-y', '--year', help='Processing year. Needed.', dest='pYear')

    # parse the given system arguments
    (options, args) = parser.parse_args()
    # do checks on the parsed options
    if (not options.path_coastal_mask) or (not os.path.isfile(os.path.normpath(options.path_coastal_mask))):
        parser.error("the -c argument for the coastal mask input raster is missing or the given path doesn't exist!!!")
    if (not options.path_in) or (not os.path.isdir(os.path.normpath(options.path_in))):
        parser.error("the -i argument for CLCACC input directory is missing or the given path doesn't exist!!!")
    if not options.path_out:
        parser.error("the -o argument to define output path is missing!!!")
    if (not options.pYear) or (not options.pYear.isdigit()):
        parser.error("the -y argument must be a year.")

    if len(args) != 0:
        parser.error('too many arguments')

    return options

if __name__ == '__main__':
    # hack to get the proper logger when running the script directly
    logger = logging.getLogger('inca.common.preprocessing.preprocessing_clcacc')
    main()