#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
SYNOPSIS
    inca crop_pollination <config_file.yaml> --run-name <NAME> --output-dir <PATH> [--continue] [--verbose] [--help]

DESCRIPTION
    This module runs the INCA crop pollination calculations.

    Please use a config.yaml to define all the parameters to run the script.
    Note: if output-dir is not specified, the current working directory will be used.

PREREQUISITES
    Python => 3.7.0

EXAMPLES
    python -m inca crop_pollination '/data/inca_vol1/runs/crop_pollination_parameters_test-run.yaml'  --run-name test --verbose

EXIT STATUS
    Ctrl-C  =  cancel the script at any time

AUTHOR
    Dr. Marcel Buchhorn <marcel.buchhorn@vito.be>

LICENSE
    This script is property of VITO. Copyrights or restrictions may apply.

VERSION
    1.2 (2022-02-04)
"""

import logging
import os
import time
import operator
from functools import reduce

import inca
import inca.common.geoprocessing as geo
import inca.croppollination.processing as processing

logger = logging.getLogger(__name__)

class PollinationRun(inca.Run):
    def validate(self):
        super().validate()
        #here we can specify additional script specific checks
        for year in self.years:
            if year not in self.config['paths']['EBM']:
                raise inca.InvalidConfig(f'No annual EBM raster provided for year {year}.')
            if year not in self.config['paths']['SDM']:
                raise inca.InvalidConfig(f'No annual SDM raster provided for year {year}.')
            if year not in self.config['paths']['masking_maes']:
                raise inca.InvalidConfig(f'No annual SDM raster provided for year {year}.')
    def _start(self):
        run(self.config, self.run_dir)


class HandleConfig(object):
    """Handles all manipulation of yaml files"""

    def __init__(self, data, verbose=False):
        self.verbose = verbose
        # ini
        self.data = data

    def get(self, position):
        """Access the nested YAML object by item sequence
           so you have to know the key order in the nested dict."""
        try:
            return reduce(operator.getitem, position, self.data)
        except Exception:
            raise ValueError('This requested entry is not existing in the YAML file')

    def set(self, position, value):
        """Writes or updates a value in nested YAML object by item sequence

        Note:
            all but the last element of 'position' is needed to find
            the parent directory to add the value to, then use the last element
            to specify the key for the value

        Reminder: so to add a further layer in the nested dict just give as value a new dict."""
        try:
            self.get(position[:-1])[position[-1]] = value
        except TypeError:
            # this is needed when the subsection has no keys up to now e.g. output_paths
            # then the sub-dictonary haas first to be init
            self.get(position[:-2])[position[-2]] = {}
            self.get(position[:-1])[position[-1]] = value

    def remove(self, position):
        """ Function to completly delete an entry from the nested YAML object"""
        del self.get(position[:-1])[position[-1]]


def run(config, run_dir):
    start_time = time.time()

    # change umask driver
    try:
        old_umask = os.umask(0o000)
    except Exception:
        logger.exception('Could not change UMASK setting')
        pass

    logger.info('Reading the configuration file...')
    oConfig = yaml_reader.HandleConfig(config, verbose=config['verbose'])

    logger.debug('START OF MODULE: INCA crop pollination')
    # ini the master tags for this module run
    oConfig.metadata = geo.Metadata("Dr. Marcel Buchhorn for INCA project",
                                    'crop_pollination',
                                    oConfig.get(['paths', 'extent']))
    # here starts real work
    # Part 1: Ecosystem Service Potential
    logger.info('Calculate annual Ecosystem Service Potential')
    try:
        dSPA, dPotential, dMasterMask = processing.calculatePotential(oConfig, run_dir)
    except Exception:
        logger.exception('Error during Potential calculation')
        raise

    # Part 2: Ecosystem Service Demand
    logger.info('Calculate annual Ecosystem Service Demand')
    try:
        dDemand_total, dDemand_components, = processing.calculateDemand(oConfig, run_dir, dMasterMask)
    except Exception:
        logger.exception('Error during Demand, UseArea and Mismatch calculation')
        raise

    # Part 3:  UseArea and mismatch
    logger.info('Calculate annual Use-Area and Mismatch')
    try:
        dUseArea, dMismatch = processing.calculateUseAreaMismatch(oConfig, run_dir, dSPA, dDemand_total)
    except Exception:
        logger.exception('Error during Demand, UseArea and Mismatch calculation')
        raise

    # Part 4: Calculation of CAPRI crop-specific pollination contribution
    logger.info('Calculate annual CAPRI crop-specific pollination contribution')
    try:
        dPolliContribution, gdf_NUTS = processing.calculateCAPRIPolliContribution(oConfig, run_dir, dSPA)
    except Exception:
        logger.exception('Error during pollination contribution calculation')
        raise

    # Part 5: Calculation of use
    logger.info('Calculate annual use with reference to annual ESTAT data')
    try:
        dUse = processing.calculateUse(oConfig, run_dir, dPolliContribution)
    except Exception:
        logger.exception('Error during flow calculation')
        raise

    # Part 6: Monetary Evaluation
    logger.info('run the Monetary Evaluation')
    try:
        dMonetary_current, dMonetary_real = processing.runMonetaryEvaluation(oConfig, run_dir, dUse)
    except Exception:
        logger.exception('Error during monetary evaluation')
        raise

    # Part 7: SUT table generation
    logger.info('generation of Supply&Use tables')
    try:
        processing.runSUTgeneration(oConfig, run_dir, dUse, dMonetary_current, dMonetary_real)
    except Exception:
        logger.exception('Error during SUT generation')
        raise

    # Part 8: additional table generation
    logger.info('generation of additional reporting tables')
    try:
        processing.runTABLEgeneration(oConfig, run_dir, dUse, dMonetary_current, dMonetary_real, gdf_NUTS, dPotential, dDemand_total)
    except Exception:
        logger.exception('Error during reporting table generation')
        raise

    # Part 9: use map generation for biophysical and monetary country data
    logger.info('generation of use maps in biophysical and monetary terms')
    try:
        processing.runMAPgeneration(oConfig, run_dir, dUse, dMonetary_current, dMonetary_real, dUseArea)
    except Exception:
        logger.exception('Error during use map generation')
        raise

    logger.info('All processing successfully done in the crop pollination module. Have a nice day.')
    # set umask to old settings
    try:
        os.umask(old_umask)
    except Exception:
        logger.exception('!!! Could not change UMASK setting back!!!')
        pass

    logger.debug('---------------------------')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: INCA crop pollination')

