#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
SYNOPSIS
    outlier_cleaning_aact_uv01.py

DESCRIPTION
    This snippet is for cleaning outliers in the ESTAT aact_uv01 dataset.

PREREQUISITES
    Python => 3.7.0

EXIT STATUS
    Ctrl-C  =  cancel the script at any time

AUTHOR
    Dr. Marcel Buchhorn <marcel.buchhorn@vito.be>

LICENSE
    This script is property of VITO. Copyrights or restrictions may apply.
VERSION
    1.0 (2022-02-04)
"""
import inca.common.estat
import os
import pandas as pd
import numpy as np


path_original = os.path.normpath(r'/data/inca_vol1/input_data/crop_pollination/input_raw/ESTAT/aact_uv01.tsv')

def checks(lCrops, lCountries, lYears):
    # now we loop over countries and crops to compare annual prices
    counter=0
    lflaged=[]
    for pCrop in lCrops:
        ltemp = []
        for pCountry in lCountries:
            # extract in EURO
            try:
                df_EUR = df.loc[(pCrop, 'EUR_T', pCountry), lYears]
                df_NAC = df.loc[(pCrop, 'NAC_T', pCountry), lYears]
            except:
                counter += 1
                continue

            # we have to push all ZEROs to NaN to avoid wrong assumptions
            df_EUR.replace(0, np.nan, inplace=True)
            df_NAC.replace(0, np.nan, inplace=True)

            # first check if we have data
            if pd.isnull(df_EUR).all():
                continue
            else:
                if df_EUR.equals(df_NAC):
                    ltemp.append(pCountry)
                else:
                    continue
        lflaged.append([pCrop, ltemp])

    # create output dataframe
    df_results = pd.DataFrame(lflaged, columns=['crop_type_codes', 'countries'])
    return df_results, counter

#load TSv file
df = inca.common.estat.read_tsv(path_original)

#get key parameters together
lCountries = pd.read_csv(os.path.normpath(r'/data/inca_vol1/input_data/crop_pollination/input_model/Look-Up-Tables/LUT_country-codes_EU27.csv'),
                             comment='#')['NUTS_ID'].tolist()

lCrops = list(set(['01110','01120','01300','01400','01500','01000','02110','02120','03100','02200','05000','02400','06110','06120', '06130','06210','06220','06230','02100','02200','02110','02130','02120','02910','04120']))

lYears_min = [1999,2000,2001]
lYears_all = [1999,2000,2001,2005,2006,2007,2011,2012,2013,2017,2018]

#run checks
df_problems_min, counter1 = checks(lCrops, lCountries, lYears_min)

df_problems_all, counter2 = checks(lCrops, lCountries, lYears_all)

'''
NOTE: due to the huge amount of issues we will just set all values for BELGIUM and LOUXEMBOURG to nan and let the rest as is.
'''

path_out = os.path.normpath(r'/data/inca_vol1/input_data/crop_pollination/input_model/ESTAT_data/aact_uv01_adapted.tsv')

for pCrop in lCrops:
    for pCountry in ['BE', 'LU']:
        #reset all values to nan
        try:
            df.loc[(pCrop, 'EUR_T', pCountry), :] = np.nan
        except:
            continue

inca.common.estat.write_tsv(df, path_out)
