#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
FUNCTIONS and CLASSES to process the data in the crop pollination model

DESCRIPTION
    Various helper functions which are used in all kinds of modules
    during the crop pollination module.

PREREQUISITES
    Python => 3.7.0

AUTHOR
    Dr. Marcel Buchhorn <marcel.buchhorn@vito.be>

LICENSE
    This script is property of VITO. Copyrights or restrictions may apply.

VERSION
    1.3 (2022-02-04)
"""

import logging
import os
import math
import subprocess

import geopandas as gpd
import numpy as np
import pandas as pd
import rasterio
import rasterio.mask
import shapely

import inca.common.geoprocessing as geoprocessing
import inca.common.fileprocessing as fileprocessing
import inca.common.nuts

logger = logging.getLogger(__name__)

def calculatePotential(oConfig, run_dir, block_shape=(256, 256)):
    ''' function to calculate the annual Ecosystem Service Potential in crop pollination'''
    # ini dict for annual output files
    dPotential = {}
    dSPA = {}
    dMasterMask = {}
    # create output root
    out_path_root = os.path.join(run_dir, 'results')
    os.makedirs(out_path_root, exist_ok=True)
    out_path_root2 = os.path.join(run_dir, 'additional_results')
    os.makedirs(out_path_root2, exist_ok=True)
    out_path_root3 = os.path.join(run_dir, 'temp')
    os.makedirs(out_path_root3, exist_ok=True)

    ##get all input together
    #dict of EBM & SDM files
    dAnnual_EBM = oConfig.get(['paths', 'EBM'])
    dAnnual_SDM = oConfig.get(['paths', 'SDM'])
    #additional masking raster file
    path_extent = os.path.normpath(oConfig.get(['paths', 'extent']))
    #LUT for category thresholds
    path_LUT_cats = os.path.normpath(oConfig.get(['paths', 'potential_class_thresholds']))
    df_LUT_cats = pd.read_csv(path_LUT_cats, comment='#', index_col='category')

    #landcover map to limit to agriculture areas
    dannual_MAES = oConfig.get(['paths', 'masking_maes'])
    pAgriculture = oConfig.get(['parameters', 'MAES_class_agriculture'])

    # start annual processing
    for pYear in oConfig.get(['years']):
        logger.info('* run processing for year: {}'.format(pYear))

        #get the annual input files
        path_EBM = dAnnual_EBM[pYear]
        path_SDM = dAnnual_SDM[pYear]
        path_maes = dannual_MAES[pYear]

        #now we have to run a check if EBM and SDm are in right extent, resolution and ESPG
        logger.info('** check of input files')
        logger.info('*** EBM')
        if not geoprocessing.raster_check(path_EBM, oConfig.metadata):
            logger.info('**** file is not matching specifications of given extent file. File is warped to specifications.')
            path_EBM = helper_warp(path_EBM, oConfig.metadata, run_dir, wResampling='average')
        else:
            logger.info('**** file OK.')
        logger.info('*** SDM')
        if not geoprocessing.raster_check(path_SDM, oConfig.metadata):
            logger.info('**** file is not matching specifications of given extent file. File is warped to specifications.')
            path_SDM = helper_warp(path_SDM, oConfig.metadata, run_dir, wResampling='bilinear', SDM_mode=True)
        else:
            logger.info('**** file OK.')

        # block processing
        logger.info('** generation of Ecosystem Service potential files')
        try:
            # create output file paths
            path_out_potential = os.path.join(out_path_root2, 'crop-pollination_map_potential_ratio_{}.tif'.format(pYear))
            path_out_potcats = os.path.join(out_path_root, 'crop-pollination_map_potential-categories_categories_{}.tif'.format(pYear))
            path_out_SPA = os.path.join(out_path_root, 'crop-pollination_map_potential-SPA_binary_{}.tif'.format(pYear))
            path_mastermask = os.path.join(out_path_root3, 'crop-pollination_agriculture-mask_{}.tif'.format(pYear))

            logger.debug('*** usage of EBM file: {}'.format(os.path.basename(path_EBM)))
            logger.debug('*** usage of SDM file: {}'.format(os.path.basename(path_SDM)))
            logger.debug('*** usage of mask file: {}'.format(os.path.basename(path_extent)))
            logger.debug('*** usage of maes to mask agriculture: {}'.format(os.path.basename(path_maes)))
            logger.debug('*** usage of LUT for potential categories file: {}'.format(os.path.basename(path_LUT_cats)))

            # set loop breaker if files already exists
            if os.path.exists(path_out_potential) and os.path.exists(path_out_potcats) and os.path.exists(path_out_SPA):
                dSPA[pYear] = path_out_SPA
                dPotential[pYear] = path_out_potcats
                dMasterMask[pYear] = path_mastermask
                continue

            # set up dst profiles and metadata
            profile_potential = oConfig.metadata.profile.copy()
            profile_potential.update(tiled=True,
                                     compress='LZW',
                                     blockysize=256,
                                     blockxsize=256,
                                     count=1,
                                     dtype=rasterio.float32,
                                     nodata=np.nan)
            profile_2 = profile_potential.copy()
            profile_2.update(dtype=rasterio.ubyte, nodata=0)

            profile_3 = profile_potential.copy()
            profile_3.update(dtype=rasterio.ubyte, nodata=255)

            # create tags for potential
            oConfig.metadata.read_raster_tags([path_EBM, path_SDM, path_extent])
            tags_potential = oConfig.metadata.prepare_raster_tags('Generation of Ecosystem Service Potential for year '
                                '{} by averaging the EBM and SDM datasets plus masking.'.format(pYear), 'ratio')

            # create tags for potential_categorized
            tags_cats = oConfig.metadata.prepare_raster_tags('Generation of Ecosystem Service Potential categories '
                                ' for year {} by averaging the EBM and SDM datasets and applying the thresholds for '
                                'suitability for pollinators.'.format(pYear), 'categories (1=None, 2=low, 3=medium, 4=high)')

            # create tags for SPA
            tags_SPA = oConfig.metadata.prepare_raster_tags('Generation of Service Providing Area for year {} '
                                'by evaluating the Ecosystem Service Potential with a threshold of {}.'.format(pYear, oConfig.get(['parameters', 'SPA_threshold'])), 'binary')

            # counter for verbose view --> non-esential code block
            if oConfig.verbose:
                max_counter = math.ceil(profile_potential['height'] * 1.0 / block_shape[0]) * math.ceil(
                    profile_potential['width'] * 1.0 / block_shape[1])
                counter = 1
                pstatus = 0

            # run block processing
            with rasterio.open(path_EBM) as src_ebm, \
                    rasterio.open(path_SDM) as src_sdm, \
                    rasterio.open(path_extent) as src_mask, \
                    rasterio.open(path_maes) as src_maes:
                # create new files
                with rasterio.open(path_out_potential, 'w', **profile_potential) as dst_pot, \
                        rasterio.open(path_out_potcats, 'w', **profile_2) as dst_cats, \
                        rasterio.open(path_out_SPA, 'w', **profile_3) as dst_spa, \
                        rasterio.open(path_mastermask, 'w', **profile_2) as dst_mastermask:
                    # set tags
                    dst_pot.update_tags(**tags_potential)
                    dst_cats.update_tags(**tags_cats)
                    dst_spa.update_tags(**tags_SPA)

                    # now iterate over the blocks of src files and process
                    for index, window in geoprocessing.block_window_generator(block_shape, profile_potential['height'],
                                                                              profile_potential['width']):
                        # read data as masked array
                        aEBM = src_ebm.read(1, window=window, masked=True)
                        aSDM = src_sdm.read(1, window=window, masked=True)
                        aMask = src_mask.read(1, window=window)
                        aMAES = src_maes.read(1, window=window)

                        # calculate final potential
                        old_warn = np.seterr(all='ignore')  # suppress floating point errors from masked values
                        aPotential = (aEBM + aSDM) / 2.0
                        np.seterr(**old_warn)
                        #apply extent masks
                        aPotential[aMask == 0] = profile_potential['nodata']
                        aPotential[(~(aMAES == pAgriculture)) & (aMask == 1) ] = profile_potential['nodata']

                        # write to disk
                        dst_pot.write(aPotential.filled(profile_potential['nodata']).astype(profile_potential['dtype']),
                                      1, window=window)

                        #now we generate the categorized version
                        aPotential = aPotential.filled(profile_potential['nodata'])
                        aCats = np.full_like(aPotential, profile_2['nodata'], dtype=rasterio.ubyte)
                        for row in df_LUT_cats.itertuples():
                            aCats[(aPotential >= row.threshold_min) & (aPotential < row.threshold_max)] = int(row.value)
                        #apply masks to be sure
                        aCats[aMask == 0] = profile_2['nodata']
                        aCats[(~(aMAES == pAgriculture)) & (aMask == 1)] = profile_2['nodata']
                        # write block out
                        dst_cats.write(aCats.astype(profile_2['dtype']), 1, window=window)

                        #finally we generate the SPA
                        aSPA = np.full_like(aPotential, 0, dtype=rasterio.ubyte)
                        aSPA[(aPotential >= oConfig.get(['parameters', 'SPA_threshold'])) & (aPotential <= 1)] = 1
                        #apply masks
                        aSPA[aMask == 0] = profile_3['nodata']
                        aSPA[(~(aMAES == pAgriculture)) & (aMask == 1)] = profile_3['nodata']
                        # write block out
                        dst_spa.write(aSPA.astype(profile_3['dtype']), 1, window=window)

                        #write out annual  agriculture master mask (agriculture class from MAES and limited to extent of AOI)
                        aMasterMask = np.full_like(aPotential, 0, dtype=rasterio.ubyte)
                        aMasterMask[(aMAES == pAgriculture) & (aMask == 1)] = 1
                        dst_mastermask.write(aMasterMask, 1, window=window)

                        # free data
                        aEBM = None
                        aSDM = None
                        aPotential = None
                        aCats = None
                        aSPA = None
                        aMask = None
                        aMAES = None
                        aMasterMask=None

                        # update print statement in verbose view  --> non-esential code block
                        if oConfig.verbose:
                            progress = (counter / max_counter) * 100
                            if (progress - pstatus) >= 5:
                                pstatus = int(progress)
                                logger.debug('{} %'.format(pstatus))
                            counter += 1
            logger.debug('done.')
        except Exception:
            # here we delete file to rerun after error
            if os.path.exists(path_out_potential):
                os.remove(path_out_potential)
            if os.path.exists(path_out_potcats):
                os.remove(path_out_potcats)
            if os.path.exists(path_out_SPA):
                os.remove(path_out_SPA)
            raise IOError('generation of potential file for year {} was not successful'.format(pYear))
        dSPA[pYear] = path_out_SPA
        dPotential[pYear] = path_out_potcats
        dMasterMask[pYear] = path_mastermask
    return dSPA, dPotential, dMasterMask

def helper_warp(path_in, metadata, run_dir, wResampling='near', SDM_mode=False):
    ''' helper function to warp crop pollination potential files to AOI'''

    # create temp_folder
    path_temp_folder = os.path.join(run_dir, 'temp')
    os.makedirs(path_temp_folder, exist_ok=True)

    #in case for SDM file we have to make sure to fill up missing areas
    if SDM_mode == True:
        #get path
        path_temp = os.path.join(os.path.normpath(path_temp_folder),
                                 os.path.splitext(os.path.basename(os.path.normpath(path_in)))[0] + r'_filled.tif')
        #run filling in 1km scale
        geoprocessing.FillHoles(path_in, path_temp, maxdistance=25, smooth=0)
        #now set the path_in to path_temp for the next processing step
        path_in = path_temp

    # create path out
    path_out = os.path.join(os.path.normpath(path_temp_folder),
                            os.path.splitext(os.path.basename(os.path.normpath(path_in)))[0] + r'_INCA_{}m_EPSG{}.tif'.format(
                                int(metadata.profile['transform'].a),
                                metadata.profile['crs'].to_epsg()))

    # run warp
    geoprocessing.Warp2AOI(metadata, path_in, path_out, wResampling=wResampling)
    return path_out

def calculateDemand(oConfig, run_dir, dMasterMask):
    ''' function to calculate the demand in the crop pollination model '''

    ###Part 1: assign the prepared demand dataset to the processing year
    logger.info('* run processing for demand - Service Demanding Areas')
    # ini dict for annual output files
    dDemand_total = {}
    dDemand_components = {}

    # create output root
    out_path_root = os.path.join(run_dir, 'results')
    os.makedirs(out_path_root, exist_ok=True)
    out_path_root2 = os.path.join(run_dir, 'additional_results')
    os.makedirs(out_path_root2, exist_ok=True)

    ##get all input together
    #dict of prepared demand
    dPrepared_demand = oConfig.get(['paths', 'demand'])

    #check if all prepared files really exist
    logger.info('** check if prepared demand files exist')
    try:
        for key in dPrepared_demand.keys():
            for ctypes in dPrepared_demand[key].keys():
                test_path = dPrepared_demand[key][ctypes]
                if os.path.exists(test_path):
                    logger.info('*** test {} file for year {}: OK'.format(ctypes, key))
                else:
                    raise FileExistsError('the required file is not existing({})'.format(test_path))
    except Exception:
        raise IOError('the required file check could not completely carried out')

    # start annual processing
    for pYear in oConfig.get(['years']):
        logger.info('** run processing for year: {}'.format(pYear))
        # here comes rule
        if pYear <= 2005:
            demand_year = 2004
        elif (pYear >= 2006) and (pYear <= 2020):
            demand_year = 2008
        else:
            raise RuntimeError('for the to process year no CAPRI year is existing or all data is too old')

        logger.debug('** using demand files of CAPRI year: {}'.format(demand_year))

        #now assign the correct dic
        dAnnualDemand = dPrepared_demand[demand_year]

        #get correct agriculture master mask
        path_mask = dMasterMask[pYear]

        #we run through the data and assign
        dTemp = {}
        for key in dAnnualDemand.keys():
            logger.info('*** run processing for dataset: {}'.format(key))
            path_in = dAnnualDemand[key]
            #copy the file and rename to the correct pattern
            if key == 'TOTAL':
                path_out = os.path.join(out_path_root, 'crop-pollination_map_demand-SDA-{}_hectare_{}.tif'.format(key, pYear))
            else:
                path_out = os.path.join(out_path_root2, 'crop-pollination_map_demand-SDA-{}_hectare_{}.tif'.format(key, pYear))

            if os.path.exists(path_out):
                pass
            else:
                try:
                    #run a gdal_calc to copy file as well as include the annualMasterMask
                    cmd = 'gdal_calc.py --calc="numpy.minimum(A*B,100)" --outfile={} -A {} -B {} --type=Float64 --creation-option="COMPRESS=LZW"'.format(path_out, path_in, path_mask)
                    subprocess.check_call(cmd, shell=True)
                except Exception:
                    raise IOError('could not generate the file: {}'.format(path_out))

            #assign to right dictonary
            if key == 'TOTAL':
                dDemand_total[pYear] = path_out
            else:
                dTemp[key] = path_out
        #now assign the components dict to final variable
        dDemand_components[pYear] = dTemp.copy()
    return dDemand_total, dDemand_components

def calculateUseAreaMismatch(oConfig, run_dir, dSPA, dDemand_total, block_shape=(256, 256)):
    ''' function to calculate the UseArea and Mismatch in the crop pollination model '''

    logger.info('* run processing for use area (Service Benefiting Areas) and mismatch')
    # ini dict for annual output files
    dUseArea = {}
    dMismatch = {}

    # create output root
    out_path_root = os.path.join(run_dir, 'additional_results')
    os.makedirs(out_path_root, exist_ok=True)

    # start annual processing
    for pYear in oConfig.get(['years']):
        logger.info('** run processing for year: {}'.format(pYear))

        #get the annual input files
        path_demand = dDemand_total[pYear]
        path_SPA = dSPA[pYear]

        # block processing
        try:
            # create output file paths
            path_out_SBA = os.path.join(out_path_root, 'crop-pollination_map_UseArea-SBA_hectare_{}.tif'.format(pYear))
            path_out_mismatch = os.path.join(out_path_root, 'crop-pollination_map_mismatch_hectare_{}.tif'.format(pYear))

            logger.debug('*** usage of SDA-total file: {}'.format(os.path.basename(path_demand)))
            logger.debug('*** usage of SPA file: {}'.format(os.path.basename(path_SPA)))

            # set loop breaker if files already exists
            if (os.path.exists(path_out_SBA)) and (os.path.exists(path_out_mismatch)):
                dUseArea[pYear] = path_out_SBA
                dMismatch[pYear] = path_out_mismatch
                continue

            # set up dst profiles and metadata
            dst_profile = oConfig.metadata.profile.copy()
            dst_profile.update(tiled=True,
                                     compress='LZW',
                                     blockysize=256,
                                     blockxsize=256,
                                     count=1,
                                     dtype=rasterio.float32,
                                     nodata=np.nan)

            # create tags
            oConfig.metadata.read_raster_tags([path_demand, path_SPA])
            tags_SBA = oConfig.metadata.prepare_raster_tags('Generation of Service Benefiting Areas for year '
                                '{} by masking the total demand (Servive Demanding Areas) with the SPA (Service Providing Areas).'.format(pYear), 'ha/km2')
            oConfig.metadata.read_raster_tags([path_demand, path_SPA])
            tags_unmet = oConfig.metadata.prepare_raster_tags('Generation of Mismatch for year '
                                '{} by masking the total demand (Servive Demanding Areas) with the SPA (Service Providing Areas).'.format(pYear), 'ha/km2')

            # counter for verbose view --> non-esential code block
            if oConfig.verbose:
                max_counter = math.ceil(dst_profile['height'] * 1.0 / block_shape[0]) * math.ceil(
                    dst_profile['width'] * 1.0 / block_shape[1])
                counter = 1
                pstatus = 0

            # run block processing
            with rasterio.open(path_demand) as src_SDA, \
                    rasterio.open(path_SPA) as src_SPA:
                # create new files
                with rasterio.open(path_out_SBA, 'w', **dst_profile) as dst_SBA, \
                        rasterio.open(path_out_mismatch, 'w', **dst_profile) as dst_mismatch:
                    # set tags
                    dst_SBA.update_tags(**tags_SBA)
                    dst_mismatch.update_tags(**tags_unmet)
                    # now iterate over the blocks of src files and process
                    for index, window in geoprocessing.block_window_generator(block_shape, dst_profile['height'],
                                                                              dst_profile['width']):
                        # read data as masked array
                        aSDA = src_SDA.read(1, window=window, masked=True)
                        aSPA = src_SPA.read(1, window=window, masked=True)

                        aMismatch = aSDA.copy()
                        aSBA = aSDA.copy()

                        #mask out area in the demand which are not in the SPA (gives: demand which is served by the providing areas) --> use areas
                        aSBA[aSPA == 0] = dst_profile['nodata']
                        #mask out area in the demand which are in the SPA (gives: demand which is not covered by the providing areas) --> unmet-demand
                        aMismatch[aSPA == 1] = dst_profile['nodata']

                        #apply extent mask - taken from SPA file
                        aSBA[aSPA.mask] = dst_profile['nodata']
                        aMismatch[aSPA.mask] = dst_profile['nodata']

                        # write to disk
                        dst_SBA.write(aSBA.filled(dst_profile['nodata']).astype(dst_profile['dtype']),
                                      1, window=window)
                        dst_mismatch.write(aMismatch.filled(dst_profile['nodata']).astype(dst_profile['dtype']),
                                      1, window=window)

                        # free data
                        aSDA = None
                        aSPA = None
                        aMismatch = None
                        aSBA = None

                        # update print statement in verbose view  --> non-esential code block
                        if oConfig.verbose:
                            progress = (counter / max_counter) * 100
                            if (progress - pstatus) >= 5:
                                pstatus = int(progress)
                                logger.debug('{} %'.format(pstatus))
                            counter += 1
            logger.debug('done.')
        except Exception:
            # here we delete file to rerun after error
            if os.path.exists(path_out_SBA):
                os.remove(path_out_SBA)
            if os.path.exists(path_out_mismatch):
                os.remove(path_out_mismatch)
            raise IOError('generation of  SBA and unmet-demand for year {} was not successful'.format(pYear))
        dUseArea[pYear] = path_out_SBA
        dMismatch[pYear] = path_out_mismatch
    return dUseArea, dMismatch

def calculateCAPRIPolliContribution(oConfig, run_dir, dSPA):
    ''' function to calculate the CAPRI crop pollination contribution - is an intermediate result and used
        to adapt the ESTAT datasets in crop production to pollination amount'''

    ###Part 1: assign the prepared CAPRI crop-specific total production to the correct processing year
    logger.info('* assignment of preprocessed CAPRI total production data to processing year')
    # ini dict for annual output files
    dAnnualProductionVRT = {}
    dAnnualProductionOrder = {}

    ##get all input together
    #dict of prepared CAPRI data
    dPrepared_data = oConfig.get(['paths', 'CAPRI_total_production'])

    #check if all prepared files really exist
    logger.info('** check if prepared CAPRI data files exist')
    try:
        for key in dPrepared_data.keys():
            for ctypes in dPrepared_data[key].keys():
                test_path = dPrepared_data[key][ctypes]
                if os.path.exists(test_path):
                    logger.info('*** test {} file for year {}: OK'.format(ctypes, key))
                else:
                    raise FileExistsError('the required file is not existing({})'.format(test_path))
    except Exception:
        raise IOError('the required file check could not completely carried out')

    # start annual processing
    for pYear in oConfig.get(['years']):
        logger.info('** run processing for year: {}'.format(pYear))
        # here comes rule
        if pYear <= 2005:
            capri_year = 2004
        elif (pYear >= 2006) and (pYear <= 2020):
            capri_year = 2008
        else:
            raise RuntimeError('for the to process year no CAPRI year is existing or all data is too old')

        logger.debug('*** using total production files of CAPRI year: {}'.format(capri_year))

        ### Part 2: create virtual raster of annual production for zonal statistics
        logger.info('*** generate temporary CAPRI total production VRT file')
        #get order of crop-types for this CAPRI year for later to assign the right columns
        lKeyProducts = list(dPrepared_data[capri_year].keys())
        #get the product paths in the right order
        lPaths = []
        for pProduct in lKeyProducts:
            lPaths.append(dPrepared_data[capri_year][pProduct])

        ## now we generate the VRT file in the Temp folder
        #first get output paths together
        path_temp_folder = os.path.join(run_dir, 'temp')
        os.makedirs(path_temp_folder, exist_ok=True)
        path_vrt = os.path.join(os.path.normpath(path_temp_folder), 'CAPRI_total_production_{}.vrt'.format(pYear))
        path_list = os.path.join(os.path.normpath(path_temp_folder), 'CAPRI_total_production_paths_{}.txt'.format(pYear))
        #create the txt file with the needed CAPRI paths in the right order
        with open(path_list, "w") as outfile:
            outfile.write("\n".join(lPaths))

        #now we run gdal VRT command
        cmd = 'gdalbuildvrt -separate -input_file_list {} -overwrite {}'.format(path_list, path_vrt)
        try:
            subprocess.check_call(cmd, shell=True)
        except Exception:
            raise FileExistsError('the needed VRT file could not be generated')

        dAnnualProductionVRT[pYear] = path_vrt
        dAnnualProductionOrder[pYear] = lKeyProducts

    ###Part3: extract the needed data of total production
    logger.info('* run zonal statistics on NUTS0')
    # ini dict for annual output files
    dPolliContribution = {}

    # create output root
    out_path_root = os.path.join(run_dir, 'intermediate_data')
    os.makedirs(out_path_root, exist_ok=True)

    ##get vector file to extract NUTS0 data
    path_vector = os.path.normpath(oConfig.get(['paths', 'NUTS']))

    # open vector file holding the regions for extraction
    gdf_NUTS = gpd.read_file(path_vector)

    # we run some filtering that we can use directly supplied NUTS-0 file without any preprocessing needed
    # first a pre-filter by a given EU country list which takes out unneeded countries at all
    # Note: that implies that it is an official NUTS file with the column 'CNTR_CODE'
    lCountries = pd.read_csv(os.path.normpath(oConfig.get(['paths', 'EU_countries'])),
                             comment='#')['NUTS_ID'].tolist()
    gdf_NUTS = gdf_NUTS[gdf_NUTS.CNTR_CODE.isin(lCountries)]

    # to run the further refinement to only select complete countries within the current extent (EU28 or specific
    # countries), we have to remove the oversea territories otherwise we are missing 5 country reports in the EU28
    # TODO: that is a hardcoded fix which will always pre-filter the reporting to the EU28 EUROPE extent!
    # change to more adaptive way.
    gdf_NUTS = gpd.clip(gdf_NUTS, inca.common.nuts.bbox_eu_main)

    # now we have to check which of the remaining NUTS regions are completly in our extent
    # e.g. only France extent was run - so the process would fail if still other valid countries are in the dataframe
    bbox = shapely.geometry.box(*oConfig.metadata.pextent)
    gdf_NUTS = gdf_NUTS[gdf_NUTS['geometry'].within(bbox)]
    logger.debug('** statistics for the following %s NUTS regions will be reported: %s',
                 gdf_NUTS.shape[0], ', '.join(gdf_NUTS.NUTS_ID.tolist()))

    ##finally get the crop-specific depependency factor to calculate the CAPRI flow out of the total production
    df_Dependency = pd.read_csv(os.path.normpath(oConfig.get(['paths', 'CAPRI_flow_dependency'])), comment='#')

    ### now we can run the annual zonal statistic
    for pYear in oConfig.get(['years']):
        logger.info('** run processing for year: {}'.format(pYear))

        ##get that annual dataset together
        #correct CAPRI data
        path_CAPRI_VRT = dAnnualProductionVRT[pYear]
        #correct order of columns
        lColumns = dAnnualProductionOrder[pYear]
        #correct SPA
        path_SPA = dSPA[pYear]

        # init result list
        results = []
        # open the needed raster datasets
        with rasterio.open(path_CAPRI_VRT) as src_CAPRI, rasterio.open(path_SPA) as src_SPA:
            # loop over all regions in the vector file
            for row in gdf_NUTS.itertuples():
                logger.debug('*** work on country: {}'.format(row.NUTS_ID))
                # use rasterio.mask to extract needed data and corresponding SPA
                aData, _ = rasterio.mask.mask(src_CAPRI, [row.geometry], crop=True, filled=False)
                aSPA, _ = rasterio.mask.mask(src_SPA, [row.geometry], crop=True, filled=False, indexes=1)

                #mask the data to only agriculture areas
                aData[:,aSPA.mask] = 0

                #calculate total_production per crop type and country
                aTotal = np.ma.masked_less(aData.filled(0), 0).sum(axis=1).sum(axis=1).filled(0)

                #calculate flow by appling SPA mask per crop type and country
                aFlow = np.ma.masked_less(aData[:,aSPA == 1].filled(0), 0).sum(axis=1).filled(0)
                #the flow has to be multiplied with the dependencies - generate array in same order
                aDepend = df_Dependency.set_index('CAPRI_code').T[lColumns].to_numpy()[0]
                aFlow = aFlow * aDepend

                #assign
                results.append([row.NUTS_ID] + aTotal.tolist() + aFlow.tolist())

                # free
                aData = None
                aSPA = None
                aTotal = None
                aFlow = None
                aDepend = None

        # run test
        if len(results) == 0:
            logger.info('For this dataset no valid statistics could be extracted.  No further processing needed.')
            continue
        else:
            ## get results into dataframe
            #get full list of generated data
            ldf_columns = ['NUTS_ID'] + ['total_{}'.format(x) for x in lColumns] + ['flow_{}'.format(x) for x in lColumns]
            df_ratio = pd.DataFrame(results, columns=ldf_columns)

        #now we have to calculate the crop-specific country ratios and fill missing with the EU average
        #TODO: the used JRC method for the EU average is wacky - they take just the average of all country values! better is to calculate it out of the EU total values!!!
        #loop over the crop_types to calculate the ratio
        for element in lColumns:
            df_ratio[element] = df_ratio['flow_{}'.format(element)] / df_ratio['total_{}'.format(element)]
        #clean up
        df_ratio = df_ratio[['NUTS_ID']+lColumns]
        #now fill with EU average
        df_ratio.fillna(df_ratio.mean(), inplace=True)

        #assign
        dPolliContribution[pYear] = df_ratio.copy()

        #write copy to disk
        df_ratio.to_csv(os.path.join(out_path_root, 'crop-pollination_intermediate_CAPRI-crop-pollination-contribution_ratio_{}.csv'.format(pYear)))
    return dPolliContribution, gdf_NUTS

def calculateUse(oConfig, run_dir, dPolliContribution):
    ''' function to run the crop pollination flow calculation out of ESTAT crop yield statistics and CAPRI crop pollination ratio
       can be split up later in sub-functions '''

    # ini dict for annual output files
    dUse = {}

    # create output root
    out_path_root = os.path.join(run_dir, 'statistics')
    os.makedirs(out_path_root, exist_ok=True)

    #get universal input together
    path_ESTAT = oConfig.get(['paths', 'ESTAT_apro-cpsh1'])
    df_ESTAT_codes = pd.read_csv(os.path.normpath(oConfig.get(['paths', 'LUT_CAPRI-apro_codes'])), comment='#')
    lCountries = pd.read_csv(os.path.normpath(oConfig.get(['paths', 'EU_countries'])),
                             comment='#')['NUTS_ID'].tolist()

    #now we can run of the years
    for pYear in oConfig.get(['years']):
        logger.info('* run processing for year: {}'.format(pYear))

        ##get need annual input data together
        # list of CAPRI crops which are in the previous calculations annual dataset
        lCrops = dPolliContribution[pYear].columns.tolist()
        lCrops = [x for x in lCrops if x != 'NUTS_ID']
        #generate out of these CAPRI crop codes the relation of needed ESTAT codes
        lESTAT_code = df_ESTAT_codes[df_ESTAT_codes.CAPRI_code.isin(lCrops)].ESTAT_code.unique().tolist()

        ##prepare the ESTAT data
        logger.info('** preparing the ESTAT TSV data')
        df_pivot = prepareESTATdata(path_ESTAT, 'PR_HU_EU', lCountries, lESTAT_code, pYear)

        ####nearly last step - calculate the sums for the CAPRI crop_types from the ESTAT crop_types
        #NOTE: merge the LUT of the CAPRI-2-ESTAT codes and group
        logger.info('** convert from ESTAT to CAPRI crop types')
        df_use = df_pivot.T.merge(df_ESTAT_codes, how='left',left_index=True, right_on='ESTAT_code')
        #now we can group by the CAPRI codes and get the final ESTAT data
        df_use = df_use.groupby('CAPRI_code').sum().T

        ### last step - calculation of the flow by multiplying with the CAPRI pollination contribution ratio
        logger.info('** calculating the actual use by taking the pollination contribution into account')
        ##merge the correct ratios to the dataframe
        df_use = df_use.merge(dPolliContribution[pYear], how='left', left_index=True, right_on='NUTS_ID', suffixes=(None, '_ratio')).set_index('NUTS_ID')
        #now multiply  and also scale since the ESTAT data is in 1000t
        for element in lCrops:
            df_use[element] = df_use[element] * df_use['{}_ratio'.format(element)] * 1000.0
        #cleanup
        df_use = df_use[lCrops].copy()
        #assign
        dUse[pYear] = df_use.copy()
        #now write out
        df_use.to_csv(os.path.join(out_path_root, 'crop-pollination_statistics_use_tonnes_{}.csv'.format(pYear)))
    return dUse

def runMonetaryEvaluation(oConfig, run_dir, df_use):
    ''' function to calculate the current and real values of crop pollination service out of
        crop production use assosiated with crop pollination and the ESTAT market prices
    '''
    # ini dict for annual output files
    dMonetary_current = {}
    dMonetary_real = {}

    # create output root
    out_path_root = os.path.join(run_dir, 'statistics')
    os.makedirs(out_path_root, exist_ok=True)

    # get universal input together
    path_ESTAT = oConfig.get(['paths', 'ESTAT_aact-uv01'])
    df_ESTAT_codes = pd.read_csv(os.path.normpath(oConfig.get(['paths', 'LUT_CAPRI-aact_codes'])), comment='#', dtype=str)
    lCountries = pd.read_csv(os.path.normpath(oConfig.get(['paths', 'EU_countries'])),
                             comment='#')['NUTS_ID'].tolist()
    #deflator table filtered to needed countries
    df_deflator = pd.read_csv(os.path.normpath(oConfig.get(['paths', 'LUT_deflator_table'])), comment='#').set_index('NUTS_ID')
    df_deflator = df_deflator[df_deflator.index.isin(lCountries)]

    #now we can run of the years
    for pYear in oConfig.get(['years']):
        logger.info('* run processing for year: {}'.format(pYear))

        ##get need annual input data together
        # list of CAPRI crops which are in the previous calculations annual dataset
        lCrops = df_use[pYear].columns.tolist()
        lCrops = [x for x in lCrops if x != 'NUTS_ID']
        #generate out of these CAPRI crop codes the relation of needed ESTAT codes
        lESTAT_code = df_ESTAT_codes[df_ESTAT_codes.CAPRI_code.isin(lCrops)].ESTAT_code.unique().tolist()

        ##prepare the ESTAT data
        logger.info('** preparing the ESTAT TSV data')
        df_pivot = prepareESTATdata(path_ESTAT, 'EUR_T', lCountries, lESTAT_code, pYear)

        ####next step - calculate the average prices for the CAPRI crop_types from the ESTAT crop_types
        #NOTE: get ugly since we have not 1:1 relation ships
        logger.info('** convert from ESTAT to CAPRI crop prices')
        #ToDo: this JRC conversion is also questionable
        #make sure all zeros become a NA since JRC let zeros out in the average calculation for each CAPRI crop type
        df_pivot.replace(0, pd.NA, inplace=True)

        #due to the n:M relationsship between the CAPRI and ESTAT we have to run a loop
        for element in lCrops:
            lUse = df_ESTAT_codes[df_ESTAT_codes.CAPRI_code == element].ESTAT_code.unique().tolist()
            df_pivot[element] = df_pivot[lUse].mean(axis=1)

        logger.info('** generate current price tables')
        df_prices_current = df_pivot[lCrops].copy()
        #### NOTE: we have also to check if we have all countries in the prices or ESTAT misses one
        lMissing = [x for x in lCountries if x not in df_prices_current.index.tolist()]
        for element in lMissing:
            df_prices_current = df_prices_current.append(pd.Series(name=element))
        #fill up missing prices with the EU average
        ##TODO: that is really tricky since JRC handle some cases specially - need universal way
        #first: JRC calculates the EU average prices ignoring ZEROS in the statistics (therefore they get a higher average as when using reported 0 prices)
        #second: every country with a reported crop production for a crop type needs a prices (if not existing then it is the EU average)
        df_prices_current.fillna(df_pivot.mean(), inplace=True)

        logger.info('** run monetary evaluation in current prices')
        #merge the current price with the physical flow
        df_prices_current = df_prices_current.merge(df_use[pYear], how='left', left_index=True, right_index=True, suffixes=(None, '_t'))
        #now multiply
        for element in lCrops:
            df_prices_current[element] = df_prices_current[element] * df_prices_current['{}_t'.format(element)]
        #cleanup
        df_prices_current = df_prices_current[lCrops].copy()
        df_prices_current.index.names = ['NUTS_ID']
        df_prices_current.sort_index(inplace=True)
        #assign
        dMonetary_current[pYear] = df_prices_current.copy()
        #now write out
        df_prices_current.to_csv(os.path.join(out_path_root, 'crop-pollination_statistics_use_EURO-current_{}.csv'.format(pYear)))

        logger.info('** run monetary evaluation in real prices for reference year 2000')
        #calculate the real prices for year 2000 with deflator for processing year
        df_prices_real = df_prices_current.multiply(df_deflator['2000']/df_deflator[str(pYear)], axis='index')
        #assign
        dMonetary_real[pYear] = df_prices_real.copy()
        #now write out
        df_prices_real.to_csv(os.path.join(out_path_root, 'crop-pollination_statistics_use_EURO-real_{}.csv'.format(pYear)))

    return dMonetary_current, dMonetary_real

def prepareESTATdata(path_ESTAT, pStrucpro, lCountries, lESTAT_code, pYear):
    '''
    helper function to prepare the ESTAT TSV datasets to be usable in crop pollination
    '''
    # first read in tsv file into dataframe
    df = pd.read_csv(os.path.normpath(path_ESTAT), sep='\t')
    # now we have to split up the first column one further
    df = df.join(
        df[df.columns[0]].str.split(",", expand=True).rename(columns={0: 'crop_type', 1: 'strucpro', 2: 'geo'}))
    df.drop(df.columns[0], axis=1, inplace=True)

    # remove the white spaces from the column names
    df.columns = df.columns.str.strip()
    # strip the white spaces from columns which represent annual data
    for i in df.columns:
        if (df[i].dtype == 'object') and (i.isdigit()):
            # remove unwanted flags which are in the TSV format just are attached as letters
            df[i] = df[i].str.replace('[a-zA-Z]$', '', regex=True)
            # now we strip the white space
            df[i] = df[i].map(str.strip)
        else:
            pass

    #### filter by countries, years, crop codes
    # filter to right strucpro
    df = df[df['strucpro'] == pStrucpro]

    # to right countries - in this case always EU27 or EU28 to garantuee the EU average
    # get the country code extracted
    df['country_code'] = df['geo'].str[:2]
    df = df[df.country_code.isin(lCountries)]

    # filter to needed crop_types by code
    # that is more tricky - we have to filter on the ESTAT codes which are need to generate the CAPRI groups
    df = df[df.crop_type.isin(lESTAT_code)]

    # now we filter to the needed years
    lYears = [int(pYear) - 1, int(pYear), int(pYear) + 1]
    lNeeded = ['crop_type', 'geo', 'country_code'] + [str(x) for x in lYears if str(x) in df.columns.tolist()]
    # small check
    if len(lNeeded) <= 3:
        logger.exception(
            'no annual data in the production statistics available. update data or chose another processing year')
        raise
    df = df[lNeeded].copy()

    # now we have to convert column data into numbers and if not valid into NULL
    # we do that after the filtering to reduce the workload!
    for element in [str(x) for x in lYears if str(x) in df.columns.tolist()]:
        df[element] = pd.to_numeric(df[element], errors='coerce', downcast='float')

    # calculating the three year average
    df['three_year_avg'] = df[[str(x) for x in lYears if str(x) in df.columns.tolist()]].mean(axis=1)

    # pivot the dataframe and bring to needed core information
    df_pivot = df.pivot(index='country_code', columns='crop_type', values='three_year_avg')

    return df_pivot

def runSUTgeneration(oConfig, run_dir, dUse, dMonetary_current, dMonetary_real):
    ''' function to generate the SUT from the biophysical and monetary results
    TODO: this version depends on the Excel TEMPLATE - only EU27 template exists up to now
    Note2: make sure the column order in your dataframe matches the column order in your template - no checks are done.
    '''

    # create output root
    out_path_root = os.path.join(run_dir, 'results')
    os.makedirs(out_path_root, exist_ok=True)

    #now we can run of the years
    for pYear in oConfig.get(['years']):
        logger.info('* run processing for year: {}'.format(pYear))

        logger.info('** physical SUT')
        #get template and output path
        path_template = os.path.normpath(oConfig.get(['paths', 'templates', 'template_physical']))
        path_out = os.path.join(out_path_root, 'crop-pollination_report_SUT-physical_tonnes_{}{}'.format(pYear, os.path.splitext(path_template)[1]))
        #fill
        try:
            fileprocessing.updateExcel(dUse[pYear], pYear, path_template, path_out)
        except Exception:
            raise IOError('could not write the given Dataframe to the Excel template')

        logger.info('** monetary SUT in current prices')
        #get template and output path
        path_template = os.path.normpath(oConfig.get(['paths', 'templates', 'template_monetary_current']))
        path_out = os.path.join(out_path_root, 'crop-pollination_report_SUT-monetary_EURO-current_{}{}'.format(pYear, os.path.splitext(path_template)[1]))
        #fill
        try:
            fileprocessing.updateExcel(dMonetary_current[pYear], pYear, path_template, path_out)
        except Exception:
            raise IOError('could not write the given Dataframe to the Excel template')

        logger.info('** monetary SUT in real prices')
        #get template and output path
        path_template = os.path.normpath(oConfig.get(['paths', 'templates', 'template_monetary_real']))
        path_out = os.path.join(out_path_root, 'crop-pollination_report_SUT-monetary_EURO-real_{}{}'.format(pYear, os.path.splitext(path_template)[1]))
        #fill
        try:
            fileprocessing.updateExcel(dMonetary_real[pYear], pYear, path_template, path_out)
        except Exception:
            raise IOError('could not write the given Dataframe to the Excel template')

def runTABLEgeneration(oConfig, run_dir, dUse, dMonetary_current, dMonetary_real, gdf_NUTS, dPotential, dDemand_total):
    ''' running additional statistics on potential and demand raster datasets and give out together with
        flow as standard tables'''

    ##first part statistic extraction for potential classes and demand
    #we go it fast track since all non-agriculture areas should be in all files already nan
    logger.info('* run zonal statistics on NUTS0 for additional statistics')

    logger.debug('** statistics for the following %s NUTS regions will be reported: %s',
                 gdf_NUTS.shape[0], ', '.join(gdf_NUTS.NUTS_ID.tolist()))
    # ini dict for annual output files
    dStatsPotential = {}
    dStatsDemand = {}

    # create output root
    out_path_root = os.path.join(run_dir, 'results')
    os.makedirs(out_path_root, exist_ok=True)
    out_path_root2 = os.path.join(run_dir, 'statistics')
    os.makedirs(out_path_root2, exist_ok=True)

    #get list of potential classes
    df_LUT_cats = pd.read_csv(os.path.normpath(oConfig.get(['paths', 'potential_class_thresholds'])), comment='#', index_col='category')
    lcats = df_LUT_cats.index.tolist()

    ### now we can run the annual zonal statistic
    for pYear in oConfig.get(['years']):
        logger.info('** run processing for year: {}'.format(pYear))

        ##get that annual dataset together
        path_potential_classes = dPotential[pYear]
        path_demand = dDemand_total[pYear]

        # init result lists
        results_potential = []
        results_demand = []
        # open the needed raster datasets
        with rasterio.open(path_potential_classes) as src_pot, rasterio.open(path_demand) as src_demand:
            # loop over all regions in the vector file
            for row in gdf_NUTS.itertuples():
                logger.debug('*** work on country: {}'.format(row.NUTS_ID))
                # use rasterio.mask to extract needed data
                aPotClass, _ = rasterio.mask.mask(src_pot, [row.geometry], crop=True, filled=False, indexes=1)
                aDemand, _ = rasterio.mask.mask(src_demand, [row.geometry], crop=True, filled=False, indexes=1)

                #first statistics for demand which is in ha/km2 (one pixel is 1 km2)
                #we have nan even in the unmasked areas --> need to get out
                aDemand = np.nan_to_num(aDemand.filled(0))

                #need sum
                results_demand.append([row.NUTS_ID, aDemand.sum().tolist()])

                #second statistics for single potential classes which we have to loop over
                #need count to get extent per class in km2 area
                temp = []
                aPotClass = aPotClass.filled(0)
                for cat in lcats:
                    aResultPotClass = aPotClass[aPotClass == df_LUT_cats.loc[cat, 'value']].size
                    temp.append(aResultPotClass)
                results_potential.append([row.NUTS_ID] + temp)

                # free
                aPotClass = None
                aDemand = None
                aResultPotClass = None
        # run test
        if (len(results_demand) == 0) or (len(results_potential) == 0):
            logger.info('For this dataset no valid statistics could be extracted.  No further processing needed.')
            continue
        else:
            ## get results into dataframe
            df_potential = pd.DataFrame(results_potential, columns=['NUTS_ID'] + lcats).set_index('NUTS_ID')
            df_demand = pd.DataFrame(results_demand, columns=['NUTS_ID', 'cropland']).set_index('NUTS_ID')
            #clean up
            df_potential.sort_index(inplace=True)
            df_demand.sort_index(inplace=True)

        dStatsPotential[pYear] = df_potential.copy()
        dStatsDemand[pYear] = df_demand.copy()
        #write to disk
        df_potential.to_csv(os.path.join(out_path_root2, 'crop-pollination_statistics_potential-classes_km2_{}.csv'.format(pYear)))
        df_demand.to_csv(
            os.path.join(out_path_root2, 'crop-pollination_statistics_demand_hectare_{}.csv'.format(pYear)))

    ##second genrate the tables from template (as done in soil retention)
    logger.info('* table generation')
    dTables = {'potential_area-km2': dStatsPotential,
                   'demand_area-ha': dStatsDemand,
                   'use_tonnes': dUse,
                   'use_EURO-real': dMonetary_real,
                   'use_EURO-current': dMonetary_current}

    for key in dTables.keys():
        for pYear in oConfig.get(['years']):
            logger.info('** run processing for {} and year: {}'.format(key, pYear))

            path_template = os.path.normpath(oConfig.get(['paths', 'templates', 'tables']))
            path_out = os.path.join(out_path_root,
                                    'crop-pollination_table_{}_{}_{}{}'.format(key.split('_')[0], key.split('_')[1],
                                                                             pYear, os.path.splitext(path_template)[1]))
            #get dataframe
            df = dTables[key][pYear]
            #fill excel
            try:
                fileprocessing.updateExcel(df, pYear, path_template, path_out)
            except Exception:
                raise IOError('could not write the given Dataframe to the Excel template')
            #clean
            df = None

def runMAPgeneration(oConfig, run_dir, dUse, dMonetary_current, dMonetary_real, dUseArea):
    ''' function to generate the maps for use in biophysical and monetary terms based on country results
        and the UseArea proxy with LAU resolution'''
    logger.info('* prepare LAU areas shapefile')
    ##get vector file to extract LAU data
    path_vector = os.path.normpath(oConfig.get(['paths', 'LAU']))
    # open vector file holding the regions for extraction
    gdf_LAU = gpd.read_file(path_vector)
    # we run some filtering that we can use directly supplied LAU file without any preprocessing needed
    # first a pre-filter by a given EU country list which takes out unneeded countries at all
    # Note: that implies that it is an official LAU file with the column 'CNTR_CODE'
    lCountries = pd.read_csv(os.path.normpath(oConfig.get(['paths', 'EU_countries'])),
                             comment='#')['NUTS_ID'].tolist()
    gdf_LAU = gdf_LAU[gdf_LAU.CNTR_CODE.isin(lCountries)]
    # to run the further refinement to only select complete countries within the current extent (EU27 or specific
    # countries), we have to remove the oversea territories otherwise we are missing 5 country reports in the EU27
    # TODO: that is a hardcoded fix which will always pre-filter the reporting to the EU27 EUROPE extent!
    # change to more adaptive way.
    gdf_LAU = gpd.clip(gdf_LAU, inca.common.nuts.bbox_eu_main)
    # now we have to check which of the remaining NUTS regions are completly in our extent
    # e.g. only France extent was run - so the process would fail if still other valid countries are in the dataframe
    bbox = shapely.geometry.box(*oConfig.metadata.pextent)
    gdf_LAU = gdf_LAU[gdf_LAU['geometry'].within(bbox)]
    logger.debug('** statistics for %s LAU regions will be reported', gdf_LAU.shape[0])

    # create output root
    out_path_root = os.path.join(run_dir, 'results')
    os.makedirs(out_path_root, exist_ok=True)
    out_path_root2 = os.path.join(run_dir, 'statistics')
    os.makedirs(out_path_root2, exist_ok=True)
    out_path_root3 = os.path.join(run_dir, 'temp')
    os.makedirs(out_path_root3, exist_ok=True)
    #Note: the UseArea raster (SBA) is already filtered to crop areas - so we can go shortcut
    ### now we can run the annual zonal statistic
    for pYear in oConfig.get(['years']):
        logger.info('* run processing for year: {}'.format(pYear))
        logger.info('** statistics for USeArea per LAU get extracted')
        ##get that annual dataset together
        path_SBA = dUseArea[pYear]

        # init result lists
        results_SBA = []
        # open the needed raster datasets
        with rasterio.open(path_SBA) as src_sba:
            # loop over all regions in the vector file
            for row in gdf_LAU.itertuples():
                # use rasterio.mask to extract needed data
                aSBA, _ = rasterio.mask.mask(src_sba, [row.geometry], crop=True, filled=False, indexes=1)

                #first statistics for demand which is in ha/km2 (one pixel is 1 km2)
                #we have nan even in the unmasked areas --> need to get out
                aSBA = np.nan_to_num(aSBA.filled(0))

                #need sum
                results_SBA.append([row.GISCO_ID, aSBA.sum()])

                # free
                aSBA = None
        # run test
        if (len(results_SBA) == 0):
            logger.info('For this dataset no valid statistics could be extracted.  No further processing needed.')
            continue
        else:
            ## get results into dataframe
            df_results = pd.DataFrame(results_SBA, columns=['GISCO_ID', 'SBA_ha']).set_index('GISCO_ID')
            #clean up
            df_results.sort_index(inplace=True)
        #write to disk
        df_results.to_csv(os.path.join(out_path_root2, 'crop-pollination_statistics_UseAreas-SBA_ha_{}.csv'.format(pYear)))

        #calculate the LAU weightfactor for each country (LAU contribution on country result)
        logger.info('** area weight factor is calculated')
        #first calculate the  overall hectare of UseArea per country
        gdf_results = gdf_LAU.merge(df_results, how='left', left_on='GISCO_ID', right_index=True)
        gdf_results['SBA_country'] = gdf_results['SBA_ha'].groupby(gdf_results['CNTR_CODE']).transform('sum')
        #calculate LAU weight factor on the overall UseArea in the country
        gdf_results['WeightFactor'] = gdf_results.SBA_ha / gdf_results.SBA_country

        #calculate the LAU values for biobysical and monetary of USE
        logger.info('** biophysical and monetary USE value per LAU region are calculated')
        #merge in the physical and monetary CNTR results
        gdf_results = gdf_results.merge(dUse[pYear].sum(axis=1).rename('physical_CNTR'), how='left',
                                        left_on='CNTR_CODE', right_index=True)
        gdf_results = gdf_results.merge(dMonetary_current[pYear].sum(axis=1).rename('money_current_CNTR'), how='left',
                                        left_on='CNTR_CODE', right_index=True)
        gdf_results = gdf_results.merge(dMonetary_real[pYear].sum(axis=1).rename('money_real_CNTR'), how='left',
                                        left_on='CNTR_CODE', right_index=True)

        #calculate the LAU results
        # use in tonn

        gdf_results['physical'] = gdf_results.physical_CNTR * gdf_results.WeightFactor
        #use in EURO current
        gdf_results['money_c'] = gdf_results.money_current_CNTR * gdf_results.WeightFactor
        #use in EURO real
        gdf_results['money_r'] = gdf_results.money_real_CNTR * gdf_results.WeightFactor

        #write out as vector data
        path_vector = os.path.join(out_path_root, 'crop-pollination_map_use_multi_{}.shp'.format(pYear))
        gdf_results.to_file(path_vector)



        # use in tonnes per area of LAU (in km²)
        gdf_results['physical'] = gdf_results['physical'] / gdf_results.AREA_KM2
        gdf_results['money_c'] = gdf_results['money_c'] / gdf_results.AREA_KM2
        gdf_results['money_r'] = gdf_results['money_r'] / gdf_results.AREA_KM2

        #rasterize the LAU results in 1km and using the AOI raster
        logger.info('** final USE maps are created')
        #biophysical
        path_biophysical = os.path.join(out_path_root, 'crop-pollination_map_Use_tonnes_{}.tif'.format(pYear))
        geoprocessing.rasterize(oConfig.metadata, gdf_results, 'physical', path_biophysical, out_path_root3,
                                nodata_value=-9999, dtype='Float16')
        #monetary_current
        path_money_current = os.path.join(out_path_root, 'crop-pollination_map_Use_EURO-current_{}.tif'.format(pYear))
        geoprocessing.rasterize(oConfig.metadata, gdf_results, 'money_c', path_money_current, out_path_root3,
                                nodata_value=-9999, dtype='Float16')
        # monetary_real
        path_money_real = os.path.join(out_path_root, 'crop-pollination_map_Use_EURO-real_{}.tif'.format(pYear))
        geoprocessing.rasterize(oConfig.metadata, gdf_results, 'money_r', path_money_real, out_path_root3,
                                nodata_value=-9999, dtype='Float16')
