"""
This is the prototype for allocating the yield of the croptype-specific SDA to the different nesting areas in the SPA

currently we take two different pollination groups into account: short and long flying pollinators

INPUT:
 - SPA for the different pollinator groups (normally prepared by INCA tool - for this test prepared by hand:
              SPA resampled to 100m, warped to EPSG:3035 and cut to extent in folder 'reference_data' which was
              produced by INCA tool :) )
 - ecosystem types L1 --> normally prepared by INCA tool --> create from CORINE ACC layers from 2018
 - reporting shape and statistics raster for BE (NUTS-0 and NUTS-2) prepared by INCA-tool for year 2018
 - yield maps for pollinator dependant crops (currently only CAPRI classes) for year 2018 in t/ha (IMPORTANT)
 - LUT for the crop-specific pollination dependency on the yield (here for CAPRI classes)

Note: this prototype is not using any block processing. in final script we need block processing so make sure to
       load bufferes etc when running delination or yield assignment via moving window

"""
import inca
from inca.common.geoprocessing import GeoProcessing, statistics_byArea_byET, SUM
import os
import numpy as np
import pandas as pd
import rasterio
from scipy.ndimage.morphology import binary_dilation
from inca.common.ecosystem import ECO_ID, CROP, types_l1
from inca.common.nuts import NUTS_ID

# distance of the pollinator groups in metre
DISTANCE_LONG = 1000
DISTANCE_SHORT = 200

POLLINATORS = {'short': DISTANCE_SHORT,
               'long': DISTANCE_LONG}

##### PARAMETER section (later via tool or config file)
# Helper - later not in config
path_base = os.path.normpath(r'\\netapp03.vgt.vito.be\inca_vol1\test\Marcel\crop_pollination_prototype')
path_SPA_short = os.path.join(path_base, r'input_data\SPA_pre-processed\SPA_short-distance_2018.tif')
path_SPA_long = os.path.join(path_base, r'input_data\SPA_pre-processed\SPA_long-distance_2018.tif')

# that would come from config file or previous processing steps
pYear = 2018
paths_SPA = {'short': path_SPA_short, 'long': path_SPA_long}
path_LUT_contribution = os.path.join(path_base, r'input_data\reference_data\LUT_CAPRI-pollination_crop-dependencies.csv')
path_yields = {'APPL': os.path.join(path_base, r'input_data\SDA_pre-processed\YIELD_APPL_t-ha_2018_100m_EPSG3035.tif'),
               'CITR': os.path.join(path_base, r'input_data\SDA_pre-processed\YIELD_CITR_t-ha_2018_100m_EPSG3035.tif'),
               'OFRU': os.path.join(path_base, r'input_data\SDA_pre-processed\YIELD_OFRU_t-ha_2018_100m_EPSG3035.tif'),
               'OOIL': os.path.join(path_base, r'input_data\SDA_pre-processed\YIELD_OOIL_t-ha_2018_100m_EPSG3035.tif'),
               'PULS': os.path.join(path_base, r'input_data\SDA_pre-processed\YIELD_PULS_t-ha_2018_100m_EPSG3035.tif'),
               'RAPE': os.path.join(path_base, r'input_data\SDA_pre-processed\YIELD_RAPE_t-ha_2018_100m_EPSG3035.tif'),
               'SOYA': os.path.join(path_base, r'input_data\SDA_pre-processed\YIELD_SOYA_t-ha_2018_100m_EPSG3035.tif'),
               'SUNF': os.path.join(path_base, r'input_data\SDA_pre-processed\YIELD_SUNF_t-ha_2018_100m_EPSG3035.tif'),
               'TEXT': os.path.join(path_base, r'input_data\SDA_pre-processed\YIELD_TEXT_t-ha_2018_100m_EPSG3035.tif'),
               'TOMA': os.path.join(path_base, r'input_data\SDA_pre-processed\YIELD_TOMA_t-ha_2018_100m_EPSG3035.tif')}

# stuff generated by the INCA tool
path_eco_L1 = os.path.join(path_base, r'input_data\reference_data\ecosystem_types_L1_2008.tif')
path_statistic_raster = os.path.join(path_base, r'input_data\reference_data\statistics_shape_rasterized.tif')
path_reporting_raster = os.path.join(path_base, r'input_data\reference_data\reporting_shape_rasterized.tif')
path_out_root = os.path.normpath(r'\\netapp03.vgt.vito.be\inca_vol1\test\Marcel\crop_pollination_prototype\output')
os.makedirs(path_out_root, exist_ok=True)
##### END OF PARAMETERS

###pre-main
# init the accord object (in tool normally done by template and __init__.py
accord = GeoProcessing('Dr. Marcel Buchhorn', 'crop pollination prototype', os.path.join(path_base, 'temp2'),
                       ProjectExtentTiff=path_eco_L1)

#### processing functions for MAIN
def footprint_disc(radius):
    """ round dilineation kernel
    could be replaced with scikit-image if we can use it in INCA-tool
    """
    kernel = np.zeros((2 * radius + 1, 2 * radius + 1))
    y, x = np.ogrid[-radius:radius + 1, -radius:radius + 1]
    mask = x ** 2 + y ** 2 <= radius ** 2
    kernel[mask] = 1
    return kernel

def delinate_Pollination_Areas(group, path_SPA, path_eco_L1, path_extent, pYear):
    # calculate kernel size in pixel from distance --> currently only works for square pixels and projected coordinate systems
    radius = np.round(POLLINATORS[group] / np.sqrt(accord.pixel_area_m2()), 0)

    with rasterio.open(path_SPA) as src:
        aData = src.read(1)

    # create binary array for fast dilination
    aSPA = np.zeros_like(aData, dtype=np.bool)
    aSPA[aData == 1] = True

    # buffer
    aBuffer = binary_dilation(aSPA, footprint_disc(int(radius))).astype(np.float32)

    # now we limit to areas which can supply (here only agriculture areas) in statistical areas
    with rasterio.open(path_eco_L1) as lc, rasterio.open(path_extent) as extent:
        aLC = lc.read(1)
        mask = np.isin(aLC, types_l1[types_l1.name.isin([CROP])][ECO_ID].tolist())
        aExtent = extent.read(1)
        mask[aExtent == 0] = False

    aBuffer[~mask] = 0

    #write to disk
    path_out = os.path.join(accord.temp_dir, f'SBA_pollination-area_{group}-flying-pollinators_{pYear}.tif')
    dst_profile = accord.ref_profile.copy()
    dst_profile.update(dtype=rasterio.float32, nodata=np.nan)

    with rasterio.open(path_out, 'w', **dst_profile) as dst:
        accord.metadata.update_dataset_tags(dst, f'areas the {element} flying pollinators can reach in the SDA.',
                                            'bool', path_SPA, path_eco_L1, path_extent)
        dst.write(aBuffer, 1)
    return path_out

def usable_crop_production(crop_type, path_in, ratio, pYear, path_eco_L1, path_extent):
    """ converts the yield in t/ha into absolute production per pixel and apply the pollination ratio
    Note: we also mask out non statistical areas and non-agriculture areas - extent is base
    """
    # maps are in t/ha --> need pixel_size in ha
    px_factor = accord.pixel_area_m2() / 10000.

    path_out = os.path.join(accord.temp_dir, f'harvest-share-pollinated_{crop_type}_tonnes_{pYear}.tif')

    with rasterio.open(path_in) as src:
        aData = src.read(1, masked=True)

        dst_profile = accord.ref_profile.copy()
        dst_profile.update(dtype=src.profile['dtype'], nodata=src.nodata)

    # convert from t/ha to absolute and apply pollination dependency factor --> harvest which is due to pollination
    aData = aData * px_factor * ratio

    # now we limit to areas which can supply (here only agriculture areas) in statistical areas
    with rasterio.open(path_eco_L1) as lc, rasterio.open(path_extent) as extent:
        aLC = lc.read(1)
        mask = np.isin(aLC, types_l1[types_l1.name.isin([CROP])][ECO_ID].tolist())
        aExtent = extent.read(1)
        mask[aExtent == 0] = False

    aData[~mask] = dst_profile['nodata']

    #write out
    with rasterio.open(path_out, 'w', **dst_profile) as dst:
        accord.metadata.update_dataset_tags(dst, f'{crop_type} harvest share in tonnes per pixel which is dependent on '
                                                 f'pollination for the year {pYear}. Dependency factor of {ratio}.',
                                            'tonnes', path_in)
        dst.write(aData.filled(dst_profile['nodata']), 1)
    return path_out


def Allocate_2_nesting_site(group, crop_type, dPollination_masks, SPA, crop_harvest, pYear, path_out_root):
    """ allocating pollination harvest share back from agriculture area to the nesting sites"""
    # create mask which specify the harvest the pollinator can reach
    if group == 'short':
        with rasterio.open(dPollination_masks['short']) as src:
            aReachable = src.read(1) == 1

    elif group == 'long':
        with rasterio.open(dPollination_masks['short']) as src:
            aPart1 = src.read(1)
        with rasterio.open(dPollination_masks['long']) as src:
            aReachable = src.read(1) == 1
        # since we running ecological niche approach - the areas of the short flying pollinators has to be subtracted
        # from the long flying ones
        aReachable[aPart1 == 1] = False
        aPart1 = None
    else:
        raise ValueError(f'this pollinator group ({group}) is currently not defined.')

    # get the crop_harvest and reduce to area which brings back to nesting sites
    with rasterio.open(crop_harvest) as src:
        aHarvest = src.read(1, masked=True)

    aHarvest = aHarvest.filled(0)
    aHarvest[~aReachable] = 0
    ToDo = aHarvest.sum()
    # free
    aReachable = None

    # specify flying pattern of pollinator back to nesting area
    flying_radius = int(np.round(POLLINATORS[group] / np.sqrt(accord.pixel_area_m2()), 0))
    aAlloPattern = footprint_disc(flying_radius).astype(np.bool)
    # reset the center pixel since that is the harvest we have to allocate and can not be a nesting site
    # ToDO: check if given pixel can be a nesting site --> if yes set next line to True
    aAlloPattern[flying_radius, flying_radius] = False

    # get the SPA area in which to look for nesting area
    with rasterio.open(SPA) as src:
        aNesting = src.read(1) == 1
        final_mask = src.read(1, masked=True).mask
    # we have to pad the nesting array so that we can search for sites with kernel
    aNesting_padded = np.pad(aNesting, flying_radius, constant_values=0)
    # free
    aNesting = None

    # now comes the allocation process
    aResult = np.zeros_like(aNesting_padded, dtype=rasterio.float32)
    # loop over harvest array
    print(f'*** we have {ToDo} tonnes to allocate for {crop_type}')
    for y in range(aHarvest.shape[0]):
        for x in range(aHarvest.shape[1]):
            # check if we have a harvest to allocate
            if aHarvest[y, x] == 0:
                continue
            # get window in nesting array
            aAllocate = aNesting_padded[y:y+(2*flying_radius)+1, x:x+(2*flying_radius)+1]

            # check how many valid nesting sites we have in flying distance
            numNesting = aAllocate[aAlloPattern].sum()

            # calculate harvest share
            share = aHarvest[y, x] / float(numNesting)

            # allocate to active nesting sites
            iValid = (aAllocate == 1) & (aAlloPattern == 1)
            aResult[y:y + (2 * flying_radius) + 1, x:x + (2 * flying_radius) + 1][iValid] += share

    # remove padding from result
    aResult = aResult[flying_radius:-flying_radius, flying_radius:-flying_radius]
    Done = aResult.sum()
    print(f'*** we have allocate {Done} tonnes of {crop_type} to nesting sites')

    # write out
    aResult[final_mask] = np.nan
    #write to disk
    # first SUPPLY
    path_out = os.path.join(path_out_root, f'crop-pollination_supply_{group}-flying-pollinators_{crop_type}_{pYear}.tif')
    dst_profile = accord.ref_profile.copy()
    dst_profile.update(dtype=rasterio.float32, nodata=np.nan)

    with rasterio.open(path_out, 'w', **dst_profile) as dst:
        accord.metadata.update_dataset_tags(dst, f'Pollination dependent harvest of {crop_type} allocated to the '
                                                 f'nesting sites of {group} flying pollinators for the year {pYear}.',
                                            'tonnes', SPA, crop_harvest, dPollination_masks['short'],
                                            dPollination_masks['long'])
        dst.write(aResult, 1)

    # now USE
    path_out2 = os.path.join(path_out_root, f'crop-pollination_use_{group}-flying-pollinators_{crop_type}_{pYear}.tif')
    dst_profile = accord.ref_profile.copy()
    dst_profile.update(dtype=rasterio.float32, nodata=np.nan)

    with rasterio.open(path_out2, 'w', **dst_profile) as dst:
        accord.metadata.update_dataset_tags(dst, f'Harvest of {crop_type} allocated to {group} flying pollinators '
                                                 f'for the year {pYear}.',
                                            'tonnes', crop_harvest, dPollination_masks['short'],
                                            dPollination_masks['long'])
        dst.write(aHarvest, 1)

    return path_out, path_out2


# ------------------- MAIN

# 1. create a mask for flying distance from SPA per pollinator group
dPollination_areas = {}
for element in POLLINATORS.keys():
    print(f'* create flying maks for {element} pollinators with {POLLINATORS[element]} m')
    dPollination_areas[element] = delinate_Pollination_Areas(element,
                                                             paths_SPA[element],
                                                             path_eco_L1,
                                                             path_statistic_raster,
                                                             pYear)

# 2. convert yield maps from t/ha into absolute yield maps (t per pixel) and multiply with crop specific
# pollination dependancy --> that is the absolute yield the pollinators can bring back to their nesting sites
df_LUT = pd.read_csv(os.path.normpath(path_LUT_contribution), comment='#').set_index('CAPRI_code')
dPollinated_harvest_share = {}
for element in path_yields.keys():
    print(f'* create absolute crop production usable by pollinators: {element}')
    dPollinated_harvest_share[element] = usable_crop_production(element,
                                                                path_yields[element],
                                                                df_LUT.loc[element, 'dependency_factor'],
                                                                pYear, path_eco_L1, path_statistic_raster)

# 3. assign for each pollinator group the crop types they pollinate
# for test all crops are pollinated by all groups
POLLINATOR_CROPS = {'short': ['APPL', 'CITR', 'OFRU', 'OOIL', 'PULS', 'RAPE', 'SOYA', 'SUNF', 'TEXT', 'TOMA'],
                    'long': ['APPL', 'CITR', 'OFRU', 'OOIL', 'PULS', 'RAPE', 'SOYA', 'SUNF', 'TEXT', 'TOMA']}

# 4. assign for each SPA pixel the yield from a reachable SDA pixel (SBA areas) of the different pollinator groups
dPollination_flow = {}
for group in POLLINATOR_CROPS.keys():
    print(f'* allocating harvest contribution to SPA for {group} flying pollinators')
    dPollination_flow[group] = {}
    for crop_type in POLLINATOR_CROPS[group]:
        print(f'** running crop type: {crop_type}')
        dPollination_flow[group][crop_type] = {}
        dPollination_flow[group][crop_type]['supply'], \
            dPollination_flow[group][crop_type]['use'] = Allocate_2_nesting_site(group,
                                                                                 crop_type,
                                                                                 dPollination_areas,
                                                                                 paths_SPA[group],
                                                                                 dPollinated_harvest_share[crop_type],
                                                                                 pYear,
                                                                                 path_out_root)

# 5. extract the use for each ecosystem type (should be the use side, since the supply comes from cropland)
dStatistics = {}
for group in POLLINATOR_CROPS.keys():
    print(f'* extracting statistics for {group} flying pollinators')
    dStatistics[group] = {}
    for crop_type in POLLINATOR_CROPS[group]:
        print(f'** running crop type: {crop_type}')
        dStatistics[group][crop_type] = {}
        # first supply
        df = statistics_byArea_byET(dPollination_flow[group][crop_type]['supply'], path_reporting_raster, {'BE': 5},
                                    path_eco_L1, types_l1.set_index('name')[ECO_ID].to_dict(), block_shape=(4096, 4096))
        dStatistics[group][crop_type]['supply'] = df[[SUM]].unstack(1)[SUM]
        # second use
        df = statistics_byArea_byET(dPollination_flow[group][crop_type]['use'], path_reporting_raster, {'BE': 5},
                                    path_eco_L1, types_l1.set_index('name')[ECO_ID].to_dict(), block_shape=(4096, 4096))
        dStatistics[group][crop_type]['use'] = df[[SUM]].unstack(1)[SUM]

        #write also to disk as csv
        dic_short = types_l1.set_index('name')['short_name'].to_dict()
        # first supply
        path_export = os.path.join(path_out_root,
                                   f'crop-pollination_statistics_supply_{group}-flying-pollinators_{crop_type}_{pYear}.csv')
        dStatistics[group][crop_type]['supply'].rename(columns=dic_short).to_csv(path_export,
                                                                                 columns=types_l1['short_name'].tolist(),
                                                                                 index_label=NUTS_ID,
                                                                                 float_format='%.3f')
        # second use
        # first supply
        path_export = os.path.join(path_out_root,
                                   f'crop-pollination_statistics_use_{group}-flying-pollinators_{crop_type}_{pYear}.csv')
        dStatistics[group][crop_type]['use'].rename(columns=dic_short).to_csv(path_export,
                                                                              columns=types_l1['short_name'].tolist(),
                                                                              index_label=NUTS_ID, float_format='%.3f')

# 6. create the SUT
# take the statistics and setup the transfer to Excel --> current template could not handle it since we have to
# report by pollinator group and crop_type per ET
"""
Note: due to the distance model the values per area for supply and use do not have to be the same!!!
How to do it: take the statistics of the supply by ET and write to the supply table
              take the statistics of the use (only agriculture ET group) and allocate it to INDUSTRY on the use site
              --> the TOTAL should match overall processed areas (in this example over Belgium).
"""


pass





