#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright (c) 2022 European Union.
#
# The tool was developed with the contribution of the Joint Research Centre of the European Commission.
#
# This program is free software: you can redistribute it and/or modify it under the terms of the European Union Public
# Licence, either version 1.2 of the License, or (at your option) any later version.
# You may not use this work except in compliance with the Licence.
#
# You may obtain a copy of the Licence at: https://joinup.ec.europa.eu/collection/eupl/eupl-guidelines-faq-infographics
#
# Unless required by applicable law or agreed to in writing, software distributed under the Licence is distributed on
# an "AS IS" basis, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#
# See the Licence for the specific language governing permissions and limitations under the Licence.


"""
SYNOPSIS
    cfactor.py -e -o -a -c- t -y -n -s -p -f -l -m [-v,--verbose] [-h,--help] [--version]

DESCRIPTION
    This snippet is for preprocessing the c-factor out of statistical data and NUTS vector files needed for the
    soil retention module.
    Note: the rasterization of the final results happen in resolution, extent and EPSG of given reference raster file.
    NOTE2: it is viewable that the NUTS-2 vector file for crop_type_area should be from the closest date before the year
    you want the statistics for [2003, 2006, 2010, 2013, 2016, 2021]. (e.g. statstics 2015==NUTS2013, 2016==NUTS2016)

PREREQUISITES
    Python => 3.7.0

EXAMPLES
    python -m inca.soilretention.preprocessing.cfactor -e <fp> -o <fp> -a <fp> -c <fp> -t <fp> -y <y> -n <fp> -s <fp>
    -p <fp> -f <fp> -l <fp> -m <fp> -v

    OR

    python3 cfactor.py -e <fp> -o <fp> -a <fp> -c <fp> -t <fp> -y <y> -n <fp> -s <fp> -p <fp> -f <fp> -l <fp> -m <fp> -v

EXIT STATUS
    Ctrl-C  =  cancel the script at any time

AUTHOR
    Dr. Marcel Buchhorn <marcel.buchhorn@vito.be>

VERSION
    1.3 (2023-03-02)
"""
################## Importing modules ##################
# standard
import logging
import optparse
import os
import math
import time

# extra official modules
import pandas as pd
import rasterio
import geopandas as gpd
import numpy as np

# own modules
import inca
import inca.common.classification as classification
import inca.common.geoprocessing
import inca.common.geoprocessing as geo
import inca.soilretention.preprocessing as preprocessing
############## END importing modules ##################

logger = logging.getLogger(preprocessing.__name__)
############## MAIN ###################################


def main():
    start_time = time.time()
    # parse the arguments
    options = parse_args()

    # ini the logging
    inca.set_up_console_logging(options.verbose)

    # change umask driver
    try:
        old_umask = os.umask(0o000)
    except Exception:
        logger.exception('Could not change UMASK setting')
        pass

    logger.info('START OF MODULE: INCA soil retention cfactor preprocessing')

    # init the metadata and raster extent of project
    accord = geo.GeoProcessing("INCA-tool",
                               "Soil Retention pre-processing",
                               os.path.normpath(options.path_out),
                               os.path.normpath(options.path_extent),
                               GDAL_verbose=options.verbose)
    os.makedirs(os.path.normpath(options.path_out), exist_ok=True)

    # run the cfactor data processing for arable land
    logger.info('generating the cfactor values for arable land')
    try:
        gdf_arable = processing_cfactor_arable(options)
    except Exception:
        logger.exception('Error during processing of the datasets to generate the cfactor for arable land')
        raise

    # rasterization of cfactor arable
    logger.info('rasterization of cfactor values for arable land to INCA extent')
    try:
        path_c_arable = rasterization_arable(options, accord, gdf_arable)
    except Exception:
        logger.exception('Error during rasterization of cfactor for arable land')
        raise

    # preparing land cover map for chunk processing
    logger.info('preparation of land cover map')
    try:
        path_landcover_adapted = accord.AutomaticBring2AOI(options.path_landcover, secure_run=True)
    except Exception:
        logger.exception('Error during processing of the land cover map')
        raise

    # run the cfactor processing for non-arable land and directly finish all up
    logger.info('generating the cfactor values for non-arable land and finalizing c-factor map for AOI')
    try:
        processing_cfactor_nonarable(options, accord, path_landcover_adapted, path_c_arable)
    except Exception:
        logger.exception('Error during processing of the datasets to generate the cfactor for non-arable land and '
                         'finalization.')
        raise

    logger.info('All processing successfully done. Have a nice day.')
    # set umask to old settings
    try:
        os.umask(old_umask)
    except Exception:
        logger.exception('!!! Could not change UMASK setting back!!!')
        pass

    logger.debug('---------------------------')
    logger.debug('TOTAL TIME IN MINUTES: %10.4f', (time.time() - start_time) / 60)
    logger.debug('END OF MODULE: INCA soil retention cfactor preprocessing')


def parse_args():
    parser = optparse.OptionParser(formatter=optparse.TitledHelpFormatter(), usage=globals()['__doc__'],
                                   version="%prog v1.3")
    parser.add_option('-v', '--verbose', action='store_true', default=False, help='verbose output')
    parser.add_option('-e', '--extent', help='Path to input raster reference file. Needed.', dest='path_extent')
    parser.add_option('-o', '--pathout', help='Path to base output folder. Needed.', dest='path_out')
    parser.add_option('-a', '--apro', help='Path to Eurostat Agriculture Production Statistic in TSV. Needed.',
                      dest='path_apro_cpshr')
    parser.add_option('-c', '--countries', help='Path to LUT for needed countries. Needed.', dest='path_LUT_countries')
    parser.add_option('-t', '--croptypes', help='Path to LUT of crop type codes and cfactors. Needed.',
                      dest='path_LUT_CropTypes')
    parser.add_option('-y', '--year', help='Processing year. Needed.', dest='pYear')
    parser.add_option('-n', '--nutscrop', help='Path to NUTS-2 shapefile for crop production statistics '
                                               '(match Eurostat extraction date.). Needed.', dest='path_NUTS2_crop')
    parser.add_option('-s', '--soil', help='Path to Eurostat Farm Survey soil cover Statistics in TSV. Needed.',
                      dest='path_ef_mp_soil')
    parser.add_option('-p', '--prac', help='Path to Eurostat Farm Survey agriculture practice Statistics in '
                                           'TSV. Needed.', dest='path_ef_mp_prac')
    parser.add_option('-f', '--fcover', help='Path to annual average FCOVER raster. Needed.', dest='path_fcover')
    parser.add_option('-l', '--landcover', help='Path to land cover dataset. Needed.', dest='path_landcover')
    parser.add_option('-m', '--cland', help='Path to LUT for cfactor of land types and reclassification. Needed.',
                      dest='path_LUT_landcover')

    # parse the given system arguments
    (options, args) = parser.parse_args()
    # do checks on the parsed options
    if (not options.path_extent) or (not os.path.isfile(os.path.normpath(options.path_extent))):
        parser.error("the -i argument for the input raster extent is missing or the given path doesn't exist!!!")
    if (not options.path_apro_cpshr) or (not os.path.isfile(os.path.normpath(options.path_apro_cpshr))):
        parser.error("the -a argument for the EuroStat statistic file is missing or the given path doesn't exist!!!")
    if (not options.path_LUT_countries) or (not os.path.isfile(os.path.normpath(options.path_LUT_countries))):
        parser.error("the -c argument for the LUT of to process countries is missing or the given path doesn't exist!!")
    if (not options.path_LUT_CropTypes) or (not os.path.isfile(os.path.normpath(options.path_LUT_CropTypes))):
        parser.error("the -t argument for the LUT of crop types is missing or the given path doesn't exist!!!")
    if (not options.pYear) or (not options.pYear.isdigit()):
        parser.error("the -y argument must be a year.")
    if (not options.path_NUTS2_crop) or (not os.path.isfile(os.path.normpath(options.path_NUTS2_crop))):
        parser.error("the -n argument for the NUTS2 vector file for crop production data is missing or the given "
                     "path doesn't exist!!!")
    if (not options.path_ef_mp_soil) or (not os.path.isfile(os.path.normpath(options.path_ef_mp_soil))):
        parser.error("the -s argument for the farm survey soil cover data is missing or the given path doesn't exist!!")
    if (not options.path_ef_mp_prac) or (not os.path.isfile(os.path.normpath(options.path_ef_mp_prac))):
        parser.error("the -p argument for the farm survey agriculture practice data is missing or the given path "
                     "doesn't exist!!!")
    if (not options.path_fcover) or (not os.path.isfile(os.path.normpath(options.path_fcover))):
        parser.error("the -f argument for the FCOVER data is missing or the given path doesn't exist!!!")
    if (not options.path_landcover) or (not os.path.isfile(os.path.normpath(options.path_landcover))):
        parser.error("the -l argument for the land cover data is missing or the given path doesn't exist!!!")
    if (not options.path_LUT_landcover) or (not os.path.isfile(os.path.normpath(options.path_LUT_landcover))):
        parser.error("the -m argument for the LUT of cfactors non-arable land is missing or the given path doesn't "
                     "exist!!!")

    if len(args) != 0:
        parser.error('too many arguments')

    return options


def processing_cfactor_nonarable(options, accord, path_landcover_adapted, path_c_arable, scale_factor=0.004,
                                 block_shape=(2048, 2048)):
    ''' function to process c-factor out of FCOVER data and land cover map.
       Also finishing up the final c-factor map.'''

    # get the dict for the reclassification
    logger.info('* reading the reclassification CSV files')
    try:
        dMask = classification.CSV_2_dict(os.path.normpath(options.path_LUT_landcover),
                                          old_class='old_class', new_class='arable')
        dCmin = classification.CSV_2_dict(os.path.normpath(options.path_LUT_landcover),
                                          old_class='old_class', new_class='c_min')
        dCMax = classification.CSV_2_dict(os.path.normpath(options.path_LUT_landcover),
                                          old_class='old_class', new_class='c_max')
    except Exception:
        logger.exception('The needed reclassification dictionary could not be generated from the given CSV file')
        raise

    # output filename
    path_out = os.path.join(os.path.normpath(options.path_out), 'cfactor{}_INCA.tif'.format(options.pYear))

    # setup dst profile and metadata
    profile_dst = accord.ref_profile.copy()
    profile_dst.update(tiled=True,
                       compress='DEFLATE',
                       blockysize=256,
                       blockxsize=256,
                       count=1,
                       dtype=rasterio.float32,
                       nodata=np.nan)

    # setup tags
    accord.metadata.read_raster_tags([path_c_arable, path_landcover_adapted, os.path.normpath(options.path_fcover)])
    tags = accord.metadata.prepare_raster_tags(
        'Annual C-factor for the year {} for calculation the soil retention in INCA.'.format(
            options.pYear), 'fraction')

    if options.verbose:
        max_counter = math.ceil(profile_dst['height'] * 1.0 / block_shape[0]) * math.ceil(
            profile_dst['width'] * 1.0 / block_shape[1])
        counter = 1
        pstatus = 0

    try:
        logger.info('* calculating c-factor for non-arable land and finalizing full cfactor map')
        # run block processing
        with rasterio.open(path_c_arable) as src_carable, \
                rasterio.open(path_landcover_adapted) as src_land, \
                rasterio.open(os.path.normpath(options.path_fcover)) as src_fcover:
            # get some needed data
            try:
                s = float(src_fcover.scales[0])
                if s != 1.:
                    scale_factor = s
            except BaseException:
                logger.info('use standard scale factor since no matching metadata.')

            # create new files
            with rasterio.open(path_out, 'w', **profile_dst) as dst:
                # set tags
                dst.update_tags(**tags)

                # now iterate over the blocks of src files and process
                for index, window in inca.common.geoprocessing.block_window_generator(block_shape,
                                                                                      profile_dst['height'],
                                                                                      profile_dst['width']):
                    # read c_arable data as masked array
                    aCArable = src_carable.read(1, window=window, masked=True)
                    # ini the result array
                    aResult = np.full(aCArable.shape, profile_dst['nodata'], dtype=np.float)

                    # since the cArable is rasterized from needed NUTS-2 regions we can use that as short cut
                    if aCArable.mask.all():
                        # write empty array out
                        dst.write(aResult.astype(profile_dst['dtype']), 1, window=window)

                        # update print statement in verbose view --> non-esential code block
                        if options.verbose:
                            progress = (counter / max_counter) * 100
                            if (progress - pstatus) >= 5:
                                pstatus = int(progress)
                                logger.debug('{} %'.format(pstatus))
                            counter += 1

                        # clean up
                        aCArable = None
                        aResult = None
                        continue

                    # now we read the other needed data (masked arrays where needed)
                    aFcover = src_fcover.read(1, window=window, masked=True) * scale_factor

                    aLand = src_land.read(1, window=window)

                    # get arable/non-arable mask
                    aMask, dMask = classification.reclassification(aLand, dMask, src_land.nodata,
                                                                   255, fverbose=False)
                    # set aLand to float to get the cmin and cmax non-arable factors
                    aLand = aLand.astype(np.float)
                    # get the minimum and maximum c-factors for non-arable land
                    aCmin, dCmin = classification.reclassification(aLand, dCmin, src_land.nodata,
                                                                   profile_dst['nodata'], fverbose=False)
                    aCmax, dCMax = classification.reclassification(aLand, dCMax, src_land.nodata,
                                                                   profile_dst['nodata'], fverbose=False)

                    # calculate non-arable
                    # Note: when no valid FCOVER data is available for a point then it is set to 0 -
                    # so maximum c value is used giving higher soil erosion in these areas
                    aResult[aMask == 0] = aCmin[aMask == 0] + ((aCmax[aMask == 0] - aCmin[aMask == 0]) *
                                                               (1. - aFcover[aMask == 0].filled(0)))
                    # fuse in  arable - make sure nodata from arable get directly the new nodata value
                    aResult[aMask == 1] = aCArable[aMask == 1].filled(profile_dst['nodata'])

                    # fill --> only to be sure
                    if isinstance(aResult, np.ma.MaskedArray):
                        aResult = aResult.filled(profile_dst['nodata'])

                    # apply also the country mask => that is the value -1 == the masked nodata value
                    aResult[aCArable.mask] = np.nan

                    # write block out
                    dst.write(aResult.astype(profile_dst['dtype']), 1, window=window)

                    # free data
                    aResult = None
                    aLand = None
                    aFcover = None
                    aCArable = None
                    aMask = None
                    aCmin = None
                    aCman = None

                    # update print statement in verbose view --> non-esential code block
                    if options.verbose:
                        progress = (counter / max_counter) * 100
                        if (progress - pstatus) >= 5:
                            pstatus = int(progress)
                            logger.debug('{} %'.format(pstatus))
                        counter += 1
        logger.debug('done.')
    except Exception:
        # here we delete file to rerun after error
        if os.path.exists(path_out):
            os.remove(path_out)
        raise IOError('generation of cfactor raster for year {} was not successful'.format(options.pYear))


def rasterization_arable(options, accord, gdf_arable):
    ''' function to rasterize the cfactor arable to INCA extent

    Note: I'm not using the function from the geoprocessing.py due to needed adaptations
    ToDo: improve the geoprocesing.py functions for rasterization to run also stand-alone'''

    # final path_out
    path_out = os.path.join(os.path.normpath(options.path_out),
                            'Cfactor-arable_{}_INCA_{}m_EPSG{}.tif'.format(options.pYear,
                                                                           int(accord.ref_profile['transform'].a),
                                                                           accord.ref_profile['crs'].to_epsg()))
    # rasterize to correct extent
    accord.rasterize(gdf_arable, 'c_arable', path_out, nodata_value=-1, mode='statistical')

    # infuse extra metadata
    tags = accord.metadata.prepare_raster_tags(f'C-factor for arable land for the year {options.pYear}. Generated from '
                                               f'EUROSTAT Production Statistics and Farm Structure data. '
                                               f'Warped to EPSG, resolution and extent of INCA AOI', 'fraction')

    with rasterio.open(path_out, 'r+') as dst:
        dst.update_tags(**tags)

    return path_out


def processing_cfactor_arable(options):
    ''' function to generate the annual cfactor for arable land on NUTS-2'''

    # first: we get the needed crop type specific production areas in ha for specified year,
    # crop types and NUTS regions
    logger.info('* extracting crop type specific arable areas from Agriculture Production Statistics')
    df_crop_areas, lCountries, lCrops = prepare_CropTypeAreas(options, strucpro='AR', scaling=1000.)

    if options.verbose:
        logger.debug('* Overview percentage share of crop types on overall arable area in selected countries')
        df_annex3 = test_case_annex3(df_crop_areas)
        print('%share on total arable land for year {}'.format(options.pYear))
        print(df_annex3['perc_share'].sort_index().to_string())

    # second: calculate percentage share per crop type on total arable area per NUTS-2 resion
    logger.info('* calculate crop type specific %share on total arable area per NUTS-2 region')
    gdf_crop_NUTS2 = calculate_crop_shares(options, df_crop_areas, lCountries, lCrops)

    # third: calculate c_crop
    logger.info('* calculate crop cfactor for each NUTS-2 region')
    gdf_Ccrop = calculate_cfactor_crop(options, gdf_crop_NUTS2, lCrops)

    # fourth: prepare the agriculture management datasets
    logger.info('* prepare the Eurostat Farm Survey data')
    # since the current FSS data is only available for 2010 and 2016 we have to improvise
    if int(options.pYear) < 2006:
        logger.warning('No Farm Survey data is existing for this year. The C management component is set to 1')
        fManagement = None
    elif (int(options.pYear) >= 2006) and (int(options.pYear) <= 2013):
        fManagement = 2010
    elif (int(options.pYear) >= 2014) and (int(options.pYear) <= 2021):
        fManagement = 2016
    else:
        logger.warning('No Farm Survey data is existing for this year. The C management component is set to 1')
        fManagement = None

    if fManagement is not None:
        logger.debug('* FSS data of year {} used.'.format(fManagement))
        df_crop_prac = prepare_CropManagement(options, lCountries, fManagement, mode='prac')
        df_crop_soil = prepare_CropManagement(options, lCountries, fManagement, mode='soil')
    else:
        logger.debug('* no FSS data used')
    # fifth: calculate c_management
    logger.info('* calculate management cfactor for each NUTS-2 region')
    if fManagement is not None:
        # we have to run a check to verify if we have a crop vector file with NUTS_IDs after 2016
        # we check with EL12 NUTS2 region which is EL52 in 2016
        if 'EL52' in gdf_Ccrop.NUTS_ID.unique().tolist():
            adapt_NUTS_ids = True
        else:
            adapt_NUTS_ids = False
        df_Cmanagement = calculate_cfactor_management(df_crop_prac, df_crop_soil, adapt_NUTS_ids)

    # sixt: calculate final cfactor arable
    logger.info('* calculate final cfactor for arable areas in each NUTS-2 region')
    if fManagement is not None:
        gdf_Ccrop = gdf_Ccrop.merge(df_Cmanagement['c_management'], how='left', left_on='NUTS_ID', right_index=True)
        # if missing NUTS_2 region in c_management then fill with 1
        gdf_Ccrop['c_management'] = gdf_Ccrop['c_management'].fillna(1)
    else:
        gdf_Ccrop['c_management'] = 1
    # calculation of c_arable
    gdf_Ccrop['c_arable'] = gdf_Ccrop.c_crop * gdf_Ccrop.c_management

    return gdf_Ccrop[['NUTS_ID', 'c_arable', 'geometry']]


def calculate_cfactor_management(df_crop_prac, df_crop_soil, adapt_NUTS_ids):
    ''' function to calculate the management contribution to calculate the cfactor for arable land

    here we have a big issue.... The EUROSTAT and JRC Management tables use the NUTS2 regions with
    definition of 2013.... starting with 2016 we have to rename if changes happened e.g. France

    '''

    # combine both tables
    df = df_crop_prac.merge(df_crop_soil, how='outer', left_index=True, right_index=True)

    # calculate cfactor for tillage
    df_till = df.copy()
    # fill missing with 0
    df_till.fillna(0, inplace=True)
    # calculate overall area
    # here only the three components following publication that area %CV + %CSERV + %ZERo = 1
    df_till['area'] = df_till.TIL_CV + df_till.TIL_CSERV + df_till.TIL_ZERO
    # if the area is zero then JRC sets the NUTS2 result to Nan --> we just remove it for faster later processing
    df_till.drop(df_till[df_till['area'] <= 0].index, axis=0, inplace=True)
    # calculate following Equation 7 from the report
    df_till['c_till'] = (1 * df_till.TIL_CV / df_till.area) + (0.35 * df_till.TIL_CSERV / df_till.area) + \
                        (0.25 * df_till.TIL_ZERO / df_till.area)

    # calculate cfactor for residues
    df_res = df.copy()
    # fill missing with 0
    df_res.fillna(0, inplace=True)
    # calculate overall area (Note: here we use all columns)
    df_res['area'] = df_res.TIL_CV + df_res.TIL_CSERV + df_res.TIL_ZERO + df_res.SC_PLRES + df_res.SC_WNTCR
    # if the area is zero then JRC sets the NUTS2 result to Nan --> we just remove it for faster later processing
    df_res.drop(df_res[df_res['area'] <= 0].index, axis=0, inplace=True)
    # calculate following equation 8
    df_res['c_res'] = 1.0 * (0.88 * df_res.SC_PLRES / df_res.area) + (1 - (df_res.SC_PLRES / df_res.area))

    # calculate cfactor for winter crop
    df_cover = df.copy()
    # fill missing with 0
    df_cover.fillna(0, inplace=True)
    # calculate overall area (Note: here we use all columns)
    df_cover['area'] = df_cover.TIL_CV + df_cover.TIL_CSERV + df_cover.TIL_ZERO + df_cover.SC_PLRES + df_cover.SC_WNTCR
    # if the area is zero then JRC sets the NUTS2 result to Nan --> we just remove it for faster later processing
    df_cover.drop(df_cover[df_cover['area'] <= 0].index, axis=0, inplace=True)
    # calculate following equation 8
    df_cover['c_cover'] = 1.0 * (0.8 * df_cover.SC_WNTCR / df_cover.area) + (1 - (df_cover.SC_WNTCR / df_cover.area))

    # merge the results
    df = df.merge(df_till['c_till'], how='left', left_index=True, right_index=True)
    df = df.merge(df_res['c_res'], how='left', left_index=True, right_index=True)
    df = df.merge(df_cover['c_cover'], how='left', left_index=True, right_index=True)

    # calculate the c_management
    df['c_management'] = df.c_till * df.c_res * df.c_cover

    if adapt_NUTS_ids:
        # here we adapt the NUTS2 codes for some EU countries from 2013 (fixed in FSS data) to after 2016 names
        # ToDo: since we see a lot of NUTS changes over the years we have to find a better solution since we miss a
        #  lot of renaming
        dRecoding = {'FR41': 'FRF3', 'FR22': 'FRE2', 'FR24': 'FRB0', 'FRA4': 'FRY4', 'FRA5': 'FRY5', 'FRA3': 'FRY3',
                     'FR62': 'FRJ2',
                     'FR81': 'FRJ1', 'FR83': 'FRM0', 'FRA1': 'FRY1', 'FR82': 'FRL0', 'FR71': 'FRK2', 'FRA2': 'FRY2',
                     'FR53': 'FRI3',
                     'FR61': 'FRI1', 'FR52': 'FRH0', 'FR63': 'FRI2', 'FR72': 'FRK1', 'FR25': 'FRD1', 'FR30': 'FRE1',
                     'FR42': 'FRF1',
                     'FR26': 'FRC1', 'FR51': 'FRG0', 'FR43': 'FRC2', 'FR23': 'FRD2', 'FR21': 'FRF2', 'HU10': 'HU12',
                     'LT00': 'LT02',
                     'PL11': 'PL71', 'PL33': 'PL72', 'PL32': 'PL82', 'PL34': 'PL84', 'PL31': 'PL81', 'PL12': 'PL92'}
        df.rename(index=dRecoding, inplace=True)
    else:
        # adaption for France
        dRecoding = {'FR41': 'FRF3', 'FR22': 'FRE2', 'FR24': 'FRB0', 'FR94': 'FRY4', 'FR95': 'FRY5', 'FR93': 'FRY3',
                     'FR62': 'FRJ2',
                     'FR81': 'FRJ1', 'FR83': 'FRM0', 'FR91': 'FRY1', 'FR82': 'FRL0', 'FR71': 'FRK2', 'FR92': 'FRY2',
                     'FR53': 'FRI3',
                     'FR61': 'FRI1', 'FR52': 'FRH0', 'FR63': 'FRI2', 'FR72': 'FRK1', 'FR25': 'FRD1', 'FR30': 'FRE1',
                     'FR42': 'FRF1',
                     'FR26': 'FRC1', 'FR51': 'FRG0', 'FR43': 'FRC2', 'FR23': 'FRD2', 'FR21': 'FRF2'
                     }
        df.rename(index=dRecoding, inplace=True)

    return df


def prepare_CropManagement(options, lCountries, flag, mode='prac'):
    ''' function to prepare the EuroStat Farm Survey data out of original TSv data'''

    # first read in tsv file into dataframe
    if mode == 'prac':
        df = pd.read_csv(os.path.normpath(options.path_ef_mp_prac), sep='\t')
    else:
        df = pd.read_csv(os.path.normpath(options.path_ef_mp_soil), sep='\t')

    # now we have to split up the first column one further
    if mode == 'prac':
        df = df.join(df[df.columns[0]].str.split(",", expand=True).rename(columns={0: 'crop_rot', 1: 'tillage',
                                                                                   2: 'farmtype', 3: 'agrarea',
                                                                                   4: 'unit', 5: 'geo'}))
        df.drop(df.columns[0], axis=1, inplace=True)
    else:
        df = df.join(df[df.columns[0]].str.split(",", expand=True).rename(columns={0: 'crop_rot', 1: 'soil_cov',
                                                                                   2: 'farmtype', 3: 'agrarea',
                                                                                   4: 'unit', 5: 'geo'}))
        df.drop(df.columns[0], axis=1, inplace=True)

    # remove the white spaces from the column names
    df.columns = df.columns.str.strip()
    # strip the white spaces from columns which represent annual data
    for i in df.columns:
        if (df[i].dtype == 'object') and (i.isdigit()):
            # remove unwanted flags which are in the TSV format just are attached as letters
            df[i] = df[i].str.replace('[a-zA-Z]$', '', regex=True)
            # now we strip the white space
            df[i] = df[i].map(str.strip)
        else:
            pass

    # now we filter
    # filter to right coulmn data
    if mode == 'prac':
        df = df[(df['crop_rot'] == 'TOTAL') &
                (df['farmtype'] == 'TOTAL') &
                (df['agrarea'] == 'TOTAL') &
                (df['unit'] == 'HA') &
                (df['tillage'].isin(['TIL_CV', 'TIL_CSERV', 'TIL_ZERO']))]
    else:
        df = df[(df['crop_rot'] == 'TOTAL') &
                (df['farmtype'] == 'TOTAL') &
                (df['agrarea'] == 'TOTAL') &
                (df['unit'] == 'HA') &
                (df['soil_cov'].isin(['SC_WNTCR', 'SC_PLRES']))]

    # to right countries
    # get the country code extracted
    df['country_code'] = df['geo'].str[:2]
    df = df[df.country_code.isin(lCountries)]

    # now we have to convert column data into numbers and if not valid into NULL
    lYears = [x for x in df.columns.tolist() if x.isdigit()]
    # we do that after the filtering to reduce the workload!
    for element in lYears:
        df[element] = pd.to_numeric(df[element], errors='coerce', downcast='float')

    # check if we have different NUTS level in the data
    df['code_length'] = df.geo.str.len()
    # we need only NUTS-2, but if we have also NUTS-0 then we can close holes
    df_NUTS2 = df[df['code_length'] == 4].copy()

    # pivot the dataframe and bring to needed core information
    if mode == 'prac':
        df_pivot = df_NUTS2.pivot(index='geo', columns='tillage', values=str(flag))
    else:
        df_pivot = df_NUTS2.pivot(index='geo', columns='soil_cov', values=str(flag))

    return df_pivot


def calculate_cfactor_crop(options, gdf, lCrops):
    ''' function to calculate the cfactor for crop out of the %share on the total arable area
        for each croy type'''

    # FIRST extract the needed columns and convert to dataframe
    # do a real copy since we will change thinks
    df = pd.DataFrame(gdf[['NUTS_ID'] + lCrops], copy=True).set_index('NUTS_ID')
    df = df.T

    # load in the cfactor values for each crop type
    df_cfactor = pd.read_csv(os.path.normpath(options.path_LUT_CropTypes),
                             sep=';',
                             comment='#')[['short_label', 'cfactor']].set_index('short_label')

    # multiply the crop typ specific cfactor with each %share of crop type on arable land
    df = df.merge(df_cfactor, how='left', left_index=True, right_index=True)
    df = df.loc[:, df.columns != 'cfactor'].multiply(df['cfactor'], axis=0)

    # sum up the weighted cfactors
    df = df.T
    df['c_crop'] = df.sum(axis=1)

    # reassign the final cfactor to each NUTS2 region
    gdf = gdf.merge(df['c_crop'], how='left', left_on='NUTS_ID', right_index=True)

    # clean up and send back
    return gdf[['NUTS_ID', 'c_crop', 'geometry']]


def calculate_crop_shares(options, df, lCountries, lCrops):
    '''Calculate area percentages per crop type and per NUTS region.

    - Split input in NUTS levels.
    - Calculate the overall arable area per NUTS-0, NUTS-1 and NUTS-2 region dataset.
    - Calculate the crop type specific area percentage for the NUTS-0, NUTS-1 and NUTS-2 regions
      (crop type area / overall arable area).
    - Fill the missing crop type specific area percentages on NUTS-2 level with NUTS-1 data or if that doen't exists
      with NUTS-0 data for missing NUTS-2 regions.
    '''
    # split the dataframe by NUTS level
    df_NUTS0 = df[df.index.str.len() == 2].copy()
    df_NUTS1 = df[df.index.str.len() == 3].copy()
    df_NUTS2 = df[df.index.str.len() == 4].copy()
    # free
    df = None

    # calculate the overall arable area per NUTS level
    df_NUTS0['arable_area_ha'] = df_NUTS0.sum(axis=1)
    df_NUTS1['arable_area_ha'] = df_NUTS1.sum(axis=1)
    df_NUTS2['arable_area_ha'] = df_NUTS2.sum(axis=1)

    # now we have to filter out NUTS1 and NUTS2 regions where the arable_area_ha is 0
    # but first a test on NUTS-0 level
    if df_NUTS0[df_NUTS0['arable_area_ha'] <= 0].shape[0] != 0:
        logger.warning('Areas on NUTS level 0 were detected with ZERO arable area. Can have an influence on the '
                       'final c factor ')
    # we do a test and delete the rows where area is Zero
    df_NUTS1.drop(df_NUTS1[df_NUTS1['arable_area_ha'] <= 0].index, axis=0, inplace=True)
    df_NUTS2.drop(df_NUTS2[df_NUTS2['arable_area_ha'] <= 0].index, axis=0, inplace=True)

    # calclate the share of the crop type areas on the arable area
    # NUTS0
    df_NUTS0_share = df_NUTS0.loc[:, df_NUTS0.columns != 'arable_area_ha'].div(df_NUTS0['arable_area_ha'], axis=0)
    # here all NaN value have to be filled with 0
    df_NUTS0_share.fillna(0, inplace=True)

    # NUTS1
    df_NUTS1_share = df_NUTS1.loc[:, df_NUTS1.columns != 'arable_area_ha'].div(df_NUTS1['arable_area_ha'], axis=0)
    # here all NaN value have to be filled with 0
    df_NUTS1_share.fillna(0, inplace=True)

    # NUTS2
    df_NUTS2_share = df_NUTS2.loc[:, df_NUTS2.columns != 'arable_area_ha'].div(df_NUTS2['arable_area_ha'], axis=0)
    # here all NaN value have to be filled with 0
    df_NUTS2_share.fillna(0, inplace=True)

    # now we get the vector file to add the data on NUTS-2 level
    gdf = gpd.read_file(os.path.normpath(options.path_NUTS2_crop))
    # filter gdf to needed countries
    gdf = gdf[gdf['CNTR_CODE'].isin(lCountries)]
    # assign the data from the statistics of NUTS-2
    gdf = gdf.merge(df_NUTS2_share, how='left', left_on='NUTS_ID', right_index=True, indicator=True)

    # run test if we have to fill missing NUTS-2 regions
    if gdf[gdf['_merge'] == 'left_only'].shape[0] != 0:
        # filling of missing data for NUTS-2 regions
        # get the NUTS0 code in the NUTS2 table
        gdf['NUTS0_code'] = gdf['NUTS_ID'].str[:2]
        # get the NUTS1 code in the NUTS2 table
        gdf['NUTS1_code'] = gdf['NUTS_ID'].str[:3]

        # we use the df_NUTS1_share table as LUT now
        for i in lCrops:
            gdf.loc[(gdf[i].isnull()) & (gdf['_merge'] == 'left_only'), i] = gdf['NUTS1_code'].map(df_NUTS1_share[i])
            # now we fill if we still have missing with NUTS-0 data
            if gdf[(gdf[i].isnull()) & (gdf['_merge'] == 'left_only')].shape[0] != 0:
                gdf.loc[(gdf[i].isnull()) & (gdf['_merge'] == 'left_only'), i] = gdf['NUTS0_code'].map(df_NUTS0_share[i])

    return gdf


def prepare_CropTypeAreas(options, strucpro='AR', scaling=1000.):
    """function to prepare the EUROSTAT Agriculture Production Dataset (apro_cpsh1)

    opens the original .tsv dataset and filters to needed countries, crop types, structure of production,
    processing year (including moving average calculation), and scales to the right unit"""

    # first read in tsv file into dataframe
    df = pd.read_csv(os.path.normpath(options.path_apro_cpshr), sep='\t')
    # now we have to split up the first column one further
    df = df.join(df[df.columns[0]].str.split(",", expand=True).rename(columns={0: 'crop_type', 1: 'strucpro',
                                                                               2: 'geo'}))
    df.drop(df.columns[0], axis=1, inplace=True)

    # remove the white spaces from the column names
    df.columns = df.columns.str.strip()
    # strip the white spaces from columns which represent annual data
    for i in df.columns:
        if (df[i].dtype == 'object') and (i.isdigit()):
            # remove unwanted flags which are in the TSV format just are attached as letters
            df[i] = df[i].str.replace('[a-zA-Z]$', '', regex=True)
            # now we strip the white space
            df[i] = df[i].map(str.strip)
        else:
            pass

    # now we filter
    # filter to right strucpro
    df = df[df['strucpro'] == strucpro]

    # to right countries
    # get the country code extracted
    df['country_code'] = df['geo'].str[:2]
    lCountries = pd.read_csv(os.path.normpath(options.path_LUT_countries), comment='#')['NUTS_ID'].tolist()
    df = df[df.country_code.isin(lCountries)]

    # filter to needed crop_types by code
    lCrops = pd.read_csv(os.path.normpath(options.path_LUT_CropTypes), comment='#', sep=';')['code'].tolist()
    df = df[df.crop_type.isin(lCrops)]

    # now we filter to the needed years
    lYears = [int(options.pYear) - 1, int(options.pYear), int(options.pYear) + 1]
    lNeeded = ['crop_type', 'geo', 'country_code'] + [str(x) for x in lYears if str(x) in df.columns.tolist()]
    # small check
    if len(lNeeded) <= 3:
        logger.exception('no annual data in the production statistics available. update data or chose another '
                         'processing year')
        raise
    df = df[lNeeded]

    # additional filter - needed after discovery that Eurostat provide extra regions in their dataset
    # get the list of NUTS_ID
    gdf = gpd.read_file(os.path.normpath(options.path_NUTS2_crop))
    lCodes = gdf['NUTS_ID'].to_list()
    # now we generate the valid NUTS0, NUTS1 and NUTS2 code list
    lCodes = lCodes + list(set([x[:3] for x in lCodes])) + list(set([x[:2] for x in lCodes]))
    # filter by it
    df = df[df.geo.isin(lCodes)]

    # now we have to convert column data into numbers and if not valid into NULL
    # and apply the scaling factor to get areas in ha
    # we do that after the filtering to reduce the workload!
    for element in [str(x) for x in lYears if str(x) in df.columns.tolist()]:
        df[element] = pd.to_numeric(df[element], errors='coerce', downcast='float')
        df[element] = df[element] * scaling

    # calculating the average of the areas over chosen time period
    df['mean_ha_arable_land'] = df[[str(x) for x in lYears if str(x) in df.columns.tolist()]].mean(axis=1)

    # pivot the dataframe and bring to needed core information
    df_pivot = df.pivot(index='geo', columns='crop_type', values='mean_ha_arable_land')

    # give crop_type codes a clear name
    dCropTypes = pd.read_csv(os.path.normpath(options.path_LUT_CropTypes),
                             comment='#',
                             sep=';')[['code', 'short_label']].set_index('code').to_dict()['short_label']
    df_pivot.columns = df_pivot.columns.to_series().map(dCropTypes)
    lCrops = list(dCropTypes.values())

    return df_pivot, lCountries, lCrops


def test_case_annex3(df_pivot):
    ''' function to generate Annex III overview of share in % of crop types
         on the total arable land for the considered crop types of the whole area
         selected by the path_LUT_countries'''
    # first we only select the country NUTS0 statatistics from the Agriculture Production data
    # make sue the "geo" column is set as index in the dataframe
    df = df_pivot[df_pivot.index.str.len() == 2]
    # now we transpose it that the croptypes are the index
    df = df.T
    # we calculate the area per crop type over all countries
    df['sum'] = df.sum(axis=1)
    # we calculate the total arable area in the selected countries
    total_arable_area = df['sum'].sum()
    # finally we can calculate the share in % on the total arable area
    df['perc_share'] = df['sum'] / total_arable_area * 100.

    return df.sort_index()


if __name__ == '__main__':
    main()
